import logging
from itertools import count
from eth_utils import is_same_address, encode_hex
from web3 import Web3
from web3.contract import Contract

from microraiden import Client
from microraiden.client import Channel
from microraiden.utils import get_logs, sign_balance_proof, privkey_to_addr
from microraiden.exceptions import InvalidBalanceProof, NoOpenChannel, InvalidBalanceAmount
from microraiden.test.fixtures.channel_manager import start_channel_manager
from microraiden.channel_manager import ChannelManager
from microraiden.test.config import (
    RECEIVER_ETH_ALLOWANCE,
    RECEIVER_TOKEN_ALLOWANCE
)
import gevent
import pytest

log = logging.getLogger(__name__)


@pytest.fixture
def confirmed_open_channel(
        channel_manager: ChannelManager,
        client: Client,
        receiver_address: str,
        wait_for_blocks
):
    channel = client.open_channel(receiver_address, 10)
    wait_for_blocks(channel_manager.n_confirmations + 1)
    gevent.sleep(channel_manager.blockchain.poll_interval)
    assert (channel.sender, channel.block) in channel_manager.channels

    return channel


def test_channel_opening(
        client: Client,
        web3: Web3,
        make_account,
        channel_manager_contract,
        token_contract,
        mine_sync_event,
        wait_for_blocks,
        use_tester,
        state_db_path
):
    receiver1_privkey = make_account(RECEIVER_ETH_ALLOWANCE, RECEIVER_TOKEN_ALLOWANCE)
    receiver2_privkey = make_account(RECEIVER_ETH_ALLOWANCE, RECEIVER_TOKEN_ALLOWANCE)
    receiver_address = privkey_to_addr(receiver1_privkey)
    # make sure channel_manager1 is terminated properly, otherwise Blockchain will be running
    #  in the background, ruining other tests' results
    channel_manager1 = ChannelManager(
        web3,
        channel_manager_contract,
        token_contract,
        receiver1_privkey,
        n_confirmations=5,
        state_filename=state_db_path
    )
    start_channel_manager(channel_manager1, use_tester, mine_sync_event)

    channel_manager2 = ChannelManager(
        web3,
        channel_manager_contract,
        token_contract,
        receiver2_privkey,
        n_confirmations=5,
        state_filename=state_db_path
    )
    start_channel_manager(channel_manager2, use_tester, mine_sync_event)
    channel_manager1.wait_sync()
    channel_manager2.wait_sync()
    blockchain = channel_manager1.blockchain
    channel = client.open_channel(receiver_address, 10)
    # should be in unconfirmed channels
    wait_for_blocks(1)
    gevent.sleep(blockchain.poll_interval)
    assert (channel.sender, channel.block) not in channel_manager1.channels
    assert (channel.sender, channel.block) in channel_manager1.unconfirmed_channels
    channel_rec = channel_manager1.unconfirmed_channels[channel.sender, channel.block]
    assert is_same_address(channel_rec.receiver, receiver_address)
    assert is_same_address(channel_rec.sender, channel.sender)
    assert channel_rec.mtime == channel_rec.ctime

    # should be confirmed after n blocks
    wait_for_blocks(blockchain.n_confirmations)
    gevent.sleep(blockchain.poll_interval)
    assert (channel.sender, channel.block) in channel_manager1.channels
    channel_rec = channel_manager1.channels[channel.sender, channel.block]
    assert is_same_address(channel_rec.receiver, receiver_address)
    assert is_same_address(channel_rec.sender, channel.sender)
    assert channel_rec.balance == 0
    assert channel_rec.last_signature is None
    assert channel_rec.is_closed is False
    assert channel_rec.settle_timeout == -1

    # should not appear in other channel manager
    assert (channel.sender, channel.block) not in channel_manager2.channels
    assert (channel.sender, channel.block) not in channel_manager2.unconfirmed_channels
    channel_manager1.stop()
    channel_manager2.stop()


def test_close_unconfirmed_event(
        channel_manager: ChannelManager,
        client: Client,
        receiver_address: str,
        wait_for_blocks
):
    channel_manager.wait_sync()
    blockchain = channel_manager.blockchain
    # if unconfirmed channel is closed it should simply be forgotten
    channel = client.open_channel(receiver_address, 10)
    wait_for_blocks(1)
    gevent.sleep(blockchain.poll_interval)
    assert (channel.sender, channel.block) in channel_manager.unconfirmed_channels
    assert (channel.sender, channel.block) not in channel_manager.channels
    channel.close()
    wait_for_blocks(channel_manager.blockchain.n_confirmations)  # opening confirmed
    gevent.sleep(blockchain.poll_interval)
    assert (channel.sender, channel.block) not in channel_manager.unconfirmed_channels
    assert (channel.sender, channel.block) in channel_manager.channels
    wait_for_blocks(1)  # closing confirmed
    gevent.sleep(blockchain.poll_interval)
    assert (channel.sender, channel.block) not in channel_manager.unconfirmed_channels
    assert (channel.sender, channel.block) in channel_manager.channels


def test_close_confirmed_event(
        channel_manager: ChannelManager,
        confirmed_open_channel: Channel,
        web3: Web3,
        wait_for_blocks
):
    blockchain = channel_manager.blockchain
    channel_manager.wait_sync()

    confirmed_open_channel.close()
    wait_for_blocks(blockchain.n_confirmations)
    gevent.sleep(blockchain.poll_interval)

    channel_id = (confirmed_open_channel.sender, confirmed_open_channel.block)
    channel_rec = channel_manager.channels[channel_id]
    assert channel_rec.is_closed is True

    settle_block = channel_manager.channel_manager_contract.call().getChannelInfo(
        channel_rec.sender,
        channel_rec.receiver,
        channel_rec.open_block_number
    )[2]
    assert channel_rec.settle_timeout == settle_block


def test_channel_settled_event(
        channel_manager: ChannelManager,
        confirmed_open_channel: Channel,
        wait_for_blocks,
        web3: Web3
):
    blockchain = channel_manager.blockchain
    channel_manager.wait_sync()
    channel_id = (confirmed_open_channel.sender, confirmed_open_channel.block)
    confirmed_open_channel.close()
    wait_for_blocks(blockchain.n_confirmations)
    gevent.sleep(blockchain.poll_interval)
    channel_rec = channel_manager.channels[channel_id]
    wait_for_blocks(channel_rec.settle_timeout - web3.eth.blockNumber)
    gevent.sleep(blockchain.poll_interval)
    assert web3.eth.blockNumber >= channel_rec.settle_timeout
    assert channel_id in channel_manager.channels
    confirmed_open_channel.settle()
    wait_for_blocks(blockchain.n_confirmations)
    gevent.sleep(blockchain.poll_interval)
    assert channel_id not in channel_manager.channels


def test_topup(
        channel_manager: ChannelManager,
        confirmed_open_channel: Channel,
        wait_for_blocks
):
    blockchain = channel_manager.blockchain
    channel_manager.wait_sync()

    channel_id = (confirmed_open_channel.sender, confirmed_open_channel.block)
    confirmed_open_channel.topup(5)

    wait_for_blocks(1)
    gevent.sleep(blockchain.poll_interval)

    channel_rec = channel_manager.channels[channel_id]
    topup_txs = channel_rec.unconfirmed_topups
    assert len(topup_txs) == 1 and list(topup_txs.values())[0] == 5

    wait_for_blocks(channel_manager.blockchain.n_confirmations)
    gevent.sleep(blockchain.poll_interval)

    channel_rec = channel_manager.channels[channel_id]
    topup_txs = channel_rec.unconfirmed_topups
    assert len(topup_txs) == 0
    assert channel_rec.deposit == 15


def test_unconfirmed_topup(
        channel_manager: ChannelManager,
        client: Client,
        receiver_address: str,
        wait_for_blocks
):
    blockchain = channel_manager.blockchain
    channel_manager.wait_sync()
    channel = client.open_channel(receiver_address, 10)
    wait_for_blocks(1)
    gevent.sleep(blockchain.poll_interval)
    assert (channel.sender, channel.block) in channel_manager.unconfirmed_channels
    channel.topup(5)
    wait_for_blocks(channel_manager.blockchain.n_confirmations)
    gevent.sleep(blockchain.poll_interval)
    assert (channel.sender, channel.block) in channel_manager.channels
    channel_rec = channel_manager.channels[channel.sender, channel.block]
    assert channel_rec.deposit == 15


def test_payment(
        channel_manager: ChannelManager,
        confirmed_open_channel,
        receiver_address: str,
        receiver_privkey: str,
        sender_privkey: str,
        sender_address: str
):
    channel_manager.wait_sync()
    channel_id = (confirmed_open_channel.sender, confirmed_open_channel.block)
    channel_rec = channel_manager.channels[channel_id]
    assert channel_rec.last_signature is None
    assert channel_rec.balance == 0

    # valid transfer
    sig1 = encode_hex(confirmed_open_channel.create_transfer(2))
    channel_manager.register_payment(sender_address, channel_rec.open_block_number, 2, sig1)
    channel_rec = channel_manager.channels[channel_id]
    assert channel_rec.balance == 2
    assert channel_rec.last_signature == sig1

    # transfer signed with wrong private key
    invalid_sig = encode_hex(sign_balance_proof(
        receiver_privkey,  # should be sender's privkey
        channel_rec.receiver,
        channel_rec.open_block_number,
        4,
        channel_manager.channel_manager_contract.address
    ))
    with pytest.raises(InvalidBalanceProof):
        channel_manager.register_payment(sender_address, channel_rec.open_block_number, 4,
                                         invalid_sig)
    channel_rec = channel_manager.channels[channel_id]
    assert channel_rec.balance == 2
    assert channel_rec.last_signature == sig1

    # transfer to different receiver
    invalid_sig = encode_hex(sign_balance_proof(
        sender_privkey,
        sender_address,  # should be receiver's address
        channel_rec.open_block_number,
        4,
        channel_manager.channel_manager_contract.address
    ))
    with pytest.raises(InvalidBalanceProof):
        channel_manager.register_payment(sender_address, channel_rec.open_block_number, 4,
                                         invalid_sig)
    channel_rec = channel_manager.channels[channel_id]
    assert channel_rec.balance == 2
    assert channel_rec.last_signature == sig1

    # transfer negative amount
    invalid_sig = encode_hex(sign_balance_proof(
        sender_privkey,
        receiver_address,
        channel_rec.open_block_number,
        1,  # should be greater than 2
        channel_manager.channel_manager_contract.address
    ))
    with pytest.raises(InvalidBalanceAmount):
        channel_manager.register_payment(sender_address, channel_rec.open_block_number, 1,
                                         invalid_sig)
    channel_rec = channel_manager.channels[channel_id]
    assert channel_rec.balance == 2
    assert channel_rec.last_signature == sig1

    # parameters should match balance proof
    sig2 = encode_hex(confirmed_open_channel.create_transfer(2))
    with pytest.raises(NoOpenChannel):
        channel_manager.register_payment(receiver_address, channel_rec.open_block_number,
                                         4, sig2)
    with pytest.raises(NoOpenChannel):
        channel_manager.register_payment(sender_address, channel_rec.open_block_number + 1,
                                         4, sig2)
    with pytest.raises(InvalidBalanceProof):
        channel_manager.register_payment(sender_address, channel_rec.open_block_number,
                                         5, sig2)
    channel_rec = channel_manager.channels[channel_id]
    assert channel_rec.balance == 2
    assert channel_rec.last_signature == sig1
    channel_manager.register_payment(sender_address, channel_rec.open_block_number, 4, sig2)
    channel_rec = channel_manager.channels[channel_id]
    assert channel_rec.balance == 4
    assert channel_rec.last_signature == sig2

    # should transfer up to deposit
    sig3 = encode_hex(confirmed_open_channel.create_transfer(6))
    channel_manager.register_payment(sender_address, channel_rec.open_block_number, 10, sig3)
    channel_rec = channel_manager.channels[channel_id]
    assert channel_rec.balance == 10
    assert channel_rec.last_signature == sig3

    # transfer too much
    invalid_sig = encode_hex(sign_balance_proof(
        sender_privkey,
        receiver_address,
        channel_rec.open_block_number,
        12,  # should not be greater than 10
        channel_manager.channel_manager_contract.address
    ))
    with pytest.raises(InvalidBalanceProof):
        channel_manager.register_payment(sender_address, channel_rec.open_block_number, 12,
                                         invalid_sig)
    assert channel_rec.balance == 10
    assert channel_rec.last_signature == sig3


def test_challenge(
        channel_manager: ChannelManager,
        confirmed_open_channel: Channel,
        receiver_address: str,
        sender_address: str,
        wait_for_blocks,
        web3: Web3,
        client: Client
):
    blockchain = channel_manager.blockchain
    channel_id = (confirmed_open_channel.sender, confirmed_open_channel.block)
    sig = encode_hex(confirmed_open_channel.create_transfer(5))
    channel_manager.register_payment(sender_address, confirmed_open_channel.block, 5, sig)
    # hack channel to decrease balance
    confirmed_open_channel.update_balance(0)
    sig = confirmed_open_channel.create_transfer(3)
    block_before = web3.eth.blockNumber
    confirmed_open_channel.close()
    # should challenge and immediately settle
    for waited_blocks in count():
        logs = get_logs(client.context.token, 'Transfer', from_block=block_before - 1)
        if logs:
            break
        wait_for_blocks(1)
        assert waited_blocks < 10

    assert len([l for l in logs
                if is_same_address(l['args']['_to'], receiver_address) and
                l['args']['_value'] == 5]) == 1
    assert len([l for l in logs
                if is_same_address(l['args']['_to'], sender_address) and
                l['args']['_value'] == 5]) == 1
    wait_for_blocks(blockchain.n_confirmations)
    gevent.sleep(blockchain.poll_interval)
    assert channel_id not in channel_manager.channels

    # update channel state so that it will not be closed twice
    client.sync_channels()
    new_state = None
    for channel in client.channels:
        if all(channel.sender == confirmed_open_channel.sender,
               channel.receiver == confirmed_open_channel.receiver,
               channel.block == confirmed_open_channel.block):
            new_state = channel.state
    if new_state is None:
        confirmed_open_channel.state = confirmed_open_channel.State.closed
    else:
        confirmed_open_channel.state = new_state


def test_multiple_topups(
        channel_manager: ChannelManager,
        confirmed_open_channel: Channel,
        wait_for_blocks
):
    blockchain = channel_manager.blockchain
    channel_id = (confirmed_open_channel.sender, confirmed_open_channel.block)
    channel_rec = channel_manager.channels[channel_id]

    # first unconfirmed topup
    assert channel_rec.deposit == 10
    confirmed_open_channel.topup(5)
    wait_for_blocks(1)
    gevent.sleep(blockchain.poll_interval)

    channel_rec = channel_manager.channels[channel_id]
    assert len(channel_rec.unconfirmed_topups) == 1
    assert list(channel_rec.unconfirmed_topups.values()) == [5]
    assert channel_rec.deposit == 10

    # second unconfirmed_topups
    confirmed_open_channel.topup(10)
    wait_for_blocks(1)
    gevent.sleep(blockchain.poll_interval)
    channel_rec = channel_manager.channels[channel_id]
    assert len(channel_rec.unconfirmed_topups) >= 1  # equality if first is confirmed
    assert 10 in channel_rec.unconfirmed_topups.values()
    assert channel_rec.deposit in [10, 15]  # depends if first topup is confirmed or not

    # wait for confirmations
    wait_for_blocks(blockchain.n_confirmations)
    gevent.sleep(blockchain.poll_interval)
    channel_rec = channel_manager.channels[channel_id]
    assert len(channel_rec.unconfirmed_topups) == 0
    assert channel_rec.deposit == 25


def test_settlement(
        channel_manager: ChannelManager,
        confirmed_open_channel: Channel,
        receiver_address: str,
        wait_for_blocks,
        web3: Web3,
        token_contract: Contract,
        sender_address: str
):
    blockchain = channel_manager.blockchain
    channel_id = (confirmed_open_channel.sender, confirmed_open_channel.block)

    sig = encode_hex(confirmed_open_channel.create_transfer(2))
    channel_manager.register_payment(sender_address, confirmed_open_channel.block, 2, sig)

    confirmed_open_channel.close()
    wait_for_blocks(blockchain.n_confirmations)
    gevent.sleep(blockchain.poll_interval)
    block_before = web3.eth.blockNumber
    channel_rec = channel_manager.channels[channel_id]
    wait_for_blocks(channel_rec.settle_timeout - block_before)
    confirmed_open_channel.settle()

    logs = get_logs(token_contract, 'Transfer', from_block=block_before - 1)
    assert len([l for l in logs
                if is_same_address(l['args']['_to'], receiver_address) and
                l['args']['_value'] == 2]) == 1
    assert len([l for l in logs
                if is_same_address(l['args']['_to'], sender_address) and
                l['args']['_value'] == 8]) == 1

    wait_for_blocks(blockchain.n_confirmations)
    gevent.sleep(blockchain.poll_interval)
    assert channel_id not in channel_manager.channels


def test_cooperative(
        channel_manager: ChannelManager,
        confirmed_open_channel: Channel,
        receiver_address: str,
        web3: Web3,
        token_contract: Contract,
        wait_for_blocks,
        sender_address: str
):
    blockchain = channel_manager.blockchain
    channel_id = (confirmed_open_channel.sender, confirmed_open_channel.block)

    sig1 = encode_hex(confirmed_open_channel.create_transfer(5))
    channel_manager.register_payment(sender_address, confirmed_open_channel.block, 5, sig1)

    receiver_sig = channel_manager.sign_close(sender_address, confirmed_open_channel.block, 5)
    channel_rec = channel_manager.channels[channel_id]
    assert channel_rec.is_closed is True
    block_before = web3.eth.blockNumber
    confirmed_open_channel.close_cooperatively(receiver_sig)
    wait_for_blocks(blockchain.n_confirmations)
    gevent.sleep(blockchain.poll_interval)
    logs = get_logs(token_contract, 'Transfer', from_block=block_before - 1)
    assert len([l for l in logs
                if is_same_address(l['args']['_to'], receiver_address) and
                l['args']['_value'] == 5]) == 1
    assert len([l for l in logs
                if is_same_address(l['args']['_to'], sender_address) and
                l['args']['_value'] == 5]) == 1
    wait_for_blocks(blockchain.n_confirmations)
    gevent.sleep(blockchain.poll_interval)
    assert channel_id not in channel_manager.channels


def test_cooperative_wrong_balance_proof(
        channel_manager: ChannelManager,
        confirmed_open_channel: Channel,
        sender_address: str
):
    channel_id = (confirmed_open_channel.sender, confirmed_open_channel.block)
    channel_rec = channel_manager.channels[channel_id]

    sig1 = encode_hex(confirmed_open_channel.create_transfer(5))
    channel_manager.register_payment(sender_address, confirmed_open_channel.block, 5, sig1)

    sig2 = encode_hex(confirmed_open_channel.create_transfer(1))
    with pytest.raises(InvalidBalanceProof):
        channel_manager.sign_close(sender_address, confirmed_open_channel.block, sig2)
    assert channel_rec.is_closed is False


def test_balances(
        channel_manager: ChannelManager,
        confirmed_open_channel: Channel,
        wait_for_blocks,
        sender_address: str,
        use_tester: bool
):
    blockchain = channel_manager.blockchain
    initial_liquid_balance = channel_manager.get_liquid_balance()
    initial_locked_balance = channel_manager.get_locked_balance()
    if use_tester:
        assert initial_liquid_balance == 0
        assert initial_locked_balance == 0

    sig = encode_hex(confirmed_open_channel.create_transfer(5))
    channel_manager.register_payment(sender_address, confirmed_open_channel.block, 5, sig)

    assert channel_manager.get_liquid_balance() == initial_liquid_balance
    assert channel_manager.get_locked_balance() == 5

    receiver_sig = channel_manager.sign_close(sender_address, confirmed_open_channel.block, 5)
    confirmed_open_channel.close_cooperatively(receiver_sig)
    wait_for_blocks(blockchain.n_confirmations)
    gevent.sleep(blockchain.poll_interval)

    assert channel_manager.get_liquid_balance() == initial_liquid_balance + 5
    assert channel_manager.get_locked_balance() == initial_locked_balance


def test_different_receivers(
        web3: Web3,
        make_account,
        channel_manager_contract: Contract,
        token_contract: Contract,
        mine_sync_event,
        client: Client,
        sender_address: str,
        wait_for_blocks,
        use_tester: bool,
        state_db_path: str
):
    receiver1_privkey = make_account(RECEIVER_ETH_ALLOWANCE, RECEIVER_TOKEN_ALLOWANCE)
    receiver2_privkey = make_account(RECEIVER_ETH_ALLOWANCE, RECEIVER_TOKEN_ALLOWANCE)
    receiver1_address = privkey_to_addr(receiver1_privkey)
    channel_manager1 = ChannelManager(
        web3,
        channel_manager_contract,
        token_contract,
        receiver1_privkey,
        n_confirmations=5,
        state_filename=state_db_path
    )
    start_channel_manager(channel_manager1, use_tester, mine_sync_event)

    channel_manager2 = ChannelManager(
        web3,
        channel_manager_contract,
        token_contract,
        receiver2_privkey,
        n_confirmations=5,
        state_filename=state_db_path
    )
    start_channel_manager(channel_manager2, use_tester, mine_sync_event)
    channel_manager1.wait_sync()
    channel_manager2.wait_sync()
    blockchain = channel_manager1.blockchain

    assert channel_manager2.blockchain.n_confirmations == blockchain.n_confirmations
    assert channel_manager2.blockchain.poll_interval == blockchain.poll_interval

    # unconfirmed open
    channel = client.open_channel(receiver1_address, 10)
    wait_for_blocks(1)
    gevent.sleep(blockchain.poll_interval)
    assert (sender_address, channel.block) in channel_manager1.unconfirmed_channels
    assert (sender_address, channel.block) not in channel_manager2.unconfirmed_channels

    # confirmed open
    wait_for_blocks(blockchain.n_confirmations)
    gevent.sleep(blockchain.poll_interval)
    assert (sender_address, channel.block) in channel_manager1.channels
    assert (sender_address, channel.block) not in channel_manager2.channels

    # unconfirmed topup
    channel.topup(5)
    wait_for_blocks(1)
    gevent.sleep(blockchain.poll_interval)
    channel_rec = channel_manager1.channels[sender_address, channel.block]
    assert len(channel_rec.unconfirmed_topups) == 1
    assert channel_rec.deposit == 10

    # confirmed topup
    wait_for_blocks(blockchain.n_confirmations)
    gevent.sleep(blockchain.poll_interval)
    channel_rec = channel_manager1.channels[sender_address, channel.block]
    assert len(channel_rec.unconfirmed_topups) == 0
    assert channel_rec.deposit == 15

    # closing
    channel.close()
    wait_for_blocks(blockchain.n_confirmations)
    gevent.sleep(blockchain.poll_interval)
    channel_rec = channel_manager1.channels[sender_address, channel.block]
    assert channel_rec.is_closed is True

    # settlement
    block_before = web3.eth.blockNumber
    wait_for_blocks(channel_rec.settle_timeout - block_before)
    channel.settle()
    wait_for_blocks(blockchain.n_confirmations)
    gevent.sleep(blockchain.poll_interval)
    assert (sender_address, channel.block) not in channel_manager1.channels
    channel_manager1.stop()
    channel_manager2.stop()


def test_reorg(
        web3: Web3,
        channel_manager: ChannelManager,
        client: Client,
        receiver_address: str,
        wait_for_blocks,
        use_tester: bool
):
    if not use_tester:
        pytest.skip('Chain reorg tests only work in tester chain')
    wait_for_blocks(10)
    # create unconfirmed channel
    channel_manager.wait_sync()
    snapshot_id = web3.testing.snapshot()
    channel = client.open_channel(receiver_address, 10)
    wait_for_blocks(0)
    assert (channel.sender, channel.block) in channel_manager.unconfirmed_channels

    # remove unconfirmed channel opening with reorg
    web3.testing.revert(snapshot_id)
    snapshot_id = web3.testing.snapshot()
    wait_for_blocks(0)
    assert (channel.sender, channel.block) not in channel_manager.unconfirmed_channels
    web3.testing.mine(channel_manager.n_confirmations)
    assert (channel.sender, channel.block) not in channel_manager.channels

    # leave confirmed channel opening
    web3.testing.revert(snapshot_id)
    channel = client.open_channel(receiver_address, 10)
    wait_for_blocks(channel_manager.n_confirmations)
    assert (channel.sender, channel.block) in channel_manager.channels
    confirmed_snapshot_id = web3.testing.snapshot()
    wait_for_blocks(3)
    web3.testing.revert(confirmed_snapshot_id)
    assert (channel.sender, channel.block) in channel_manager.channels

    # remove unconfirmed topup
    channel = client.open_channel(receiver_address, 10)
    wait_for_blocks(channel_manager.n_confirmations)
    assert (channel.sender, channel.block) in channel_manager.channels
    topup_snapshot_id = web3.testing.snapshot()
    channel.topup(5)
    wait_for_blocks(0)
    channel_rec = channel_manager.channels[channel.sender, channel.block]
    assert len(channel_rec.unconfirmed_topups) == 1
    web3.testing.revert(topup_snapshot_id)
    wait_for_blocks(0)
    assert (channel.sender, channel.block) in channel_manager.channels
    channel_rec = channel_manager.channels[channel.sender, channel.block]
    assert len(channel_rec.unconfirmed_topups) == 0



import pytest  # noqa: F401
from _pytest.monkeypatch import MonkeyPatch
from flask import jsonify

from microraiden import DefaultHTTPClient
from microraiden.examples.eth_ticker import ETHTickerClient, ETHTickerProxy
from microraiden.proxy.paywalled_proxy import PaywalledProxy
from microraiden.proxy.resources import PaywalledProxyUrl


@pytest.mark.needs_xorg
def test_eth_ticker(
        empty_proxy: PaywalledProxy,
        default_http_client: DefaultHTTPClient,
        sender_privkey: str,
        receiver_privkey: str,
        monkeypatch: MonkeyPatch
):
    def get_patched(*args, **kwargs):
        body = {
            'mid': '682.435', 'bid': '682.18', 'ask': '682.69', 'last_price': '683.16',
            'low': '532.97', 'high': '684.0', 'volume': '724266.25906224',
            'timestamp': '1513167820.721733'
        }
        return jsonify(body)

    monkeypatch.setattr(PaywalledProxyUrl, 'get', get_patched)

    ETHTickerProxy(receiver_privkey, proxy=empty_proxy)
    ticker = ETHTickerClient(sender_privkey, httpclient=default_http_client, poll_interval=0.5)

    def post():
        ticker.close()

        assert ticker.pricevar.get() == '683.16 USD'
        client = default_http_client.client
        assert len(client.get_open_channels()) == 0
        ticker.success = True

    ticker.success = False
    ticker.root.after(1500, post)
    ticker.run()
    assert ticker.success



from itertools import cycle

import requests
from eth_utils import encode_hex
from munch import Munch

from microraiden import HTTPHeaders, Client
from microraiden.proxy.resources import Expensive
from microraiden.proxy.paywalled_proxy import PaywalledProxy
from flask import request

import logging
log = logging.getLogger(__name__)


class StaticPriceResource(Expensive):
    def get(self, url):
        return 'GET'

    def post(self, url):
        return 'POST'

    def put(self, url):
        return 'PUT'

    def delete(self, url):
        return 'DEL'


class DynamicPriceResource(Expensive):
    def get(self, url, res_id: int):
        return res_id, 200


class DynamicMethodResource(StaticPriceResource):
    def __init__(self, foo, bar=12345, *args, **kwargs):
        super().__init__(*args, **kwargs)
        assert foo == 42
        assert bar == 9814072356

    def price(self):
        prices = {
            'GET': 1,
            'POST': 2,
            'PUT': 3,
            'DELETE': 4
        }
        return prices[request.method]


def assert_method(method, url, headers, channel, expected_reply, expected_price=3):
    # test POST
    response = method(url, headers=headers)
    assert response.status_code == 402
    headers = HTTPHeaders.deserialize(response.headers)
    assert int(headers.price) == expected_price

    channel.update_balance(int(headers.sender_balance) + int(headers.price))

    headers = Munch()
    headers.balance = str(channel.balance)
    headers.balance_signature = encode_hex(channel.balance_sig)
    headers.sender_address = channel.sender
    headers.open_block = str(channel.block)
    headers = HTTPHeaders.serialize(headers)

    response = method(url, headers=headers)
    assert response.status_code == 200
    assert response.text.strip() == expected_reply


def test_static_price(
        empty_proxy: PaywalledProxy,
        api_endpoint_address: str,
        client: Client,
        wait_for_blocks
):
    proxy = empty_proxy
    endpoint_url = "http://" + api_endpoint_address

    proxy.add_paywalled_resource(StaticPriceResource, '/resource', 3)

    # test GET
    response = requests.get(endpoint_url + '/resource')
    assert response.status_code == 402
    headers = HTTPHeaders.deserialize(response.headers)
    assert int(headers.price) == 3

    channel = client.get_suitable_channel(headers.receiver_address, int(headers.price) * 4)
    wait_for_blocks(6)
    channel.update_balance(int(headers.price))

    headers = Munch()
    headers.balance = str(channel.balance)
    headers.balance_signature = encode_hex(channel.balance_sig)
    headers.sender_address = channel.sender
    headers.open_block = str(channel.block)
    headers = HTTPHeaders.serialize(headers)

    response = requests.get(endpoint_url + '/resource', headers=headers)
    assert response.status_code == 200
    assert response.text.strip() == 'GET'

    assert_method(requests.post, endpoint_url + '/resource', headers, channel, 'POST')
    assert_method(requests.put, endpoint_url + '/resource', headers, channel, 'PUT')
    assert_method(requests.delete, endpoint_url + '/resource', headers, channel, 'DEL')


def test_dynamic_price(
        empty_proxy: PaywalledProxy,
        api_endpoint_address: str,
        client: Client,
        wait_for_blocks
):
    proxy = empty_proxy
    endpoint_url = "http://" + api_endpoint_address

    price_cycle = cycle([1, 2, 3, 4, 5])
    url_to_price = {}  # type: Dict

    def price_fn():
        url = request.path
        if int(url.split('_')[-1]) == 0:
            return 0
        if url in url_to_price:
            price = url_to_price[url]
        else:
            price = next(price_cycle)
            url_to_price[url] = price
        return price

    proxy.add_paywalled_resource(
        DynamicPriceResource,
        '/resource_<int:res_id>',
        price=price_fn
    )

    response = requests.get(endpoint_url + '/resource_3')
    assert response.status_code == 402
    headers = HTTPHeaders.deserialize(response.headers)
    assert int(headers.price) == 1

    response = requests.get(endpoint_url + '/resource_5')
    assert response.status_code == 402
    headers = HTTPHeaders.deserialize(response.headers)
    assert int(headers.price) == 2

    response = requests.get(endpoint_url + '/resource_2')
    assert response.status_code == 402
    headers = HTTPHeaders.deserialize(response.headers)
    assert int(headers.price) == 3

    response = requests.get(endpoint_url + '/resource_3')
    assert response.status_code == 402
    headers = HTTPHeaders.deserialize(response.headers)
    assert int(headers.price) == 1

    response = requests.get(endpoint_url + '/resource_0')
    assert response.status_code == 200
    assert response.text.strip() == '0'

    channel = client.get_suitable_channel(headers.receiver_address, 2)
    wait_for_blocks(6)
    channel.update_balance(2)

    headers = Munch()
    headers.balance = str(channel.balance)
    headers.balance_signature = encode_hex(channel.balance_sig)
    headers.sender_address = channel.sender
    headers.open_block = str(channel.block)
    headers = HTTPHeaders.serialize(headers)

    response = requests.get(endpoint_url + '/resource_5', headers=headers)
    assert response.status_code == 200
    assert response.text.strip() == '5'


def test_method_price(
        empty_proxy: PaywalledProxy,
        api_endpoint_address: str,
        client: Client,
        wait_for_blocks
):
    proxy = empty_proxy
    endpoint_url = "http://" + api_endpoint_address

    proxy.add_paywalled_resource(
        DynamicMethodResource,
        '/resource',
        resource_class_args=(42,),
        resource_class_kwargs={'bar': 9814072356})

    # test GET
    response = requests.get(endpoint_url + '/resource')
    assert response.status_code == 402
    headers = HTTPHeaders.deserialize(response.headers)
    assert int(headers.price) == 1

    channel = client.get_suitable_channel(headers.receiver_address, 1 + 2 + 3 + 4)
    wait_for_blocks(6)
    channel.update_balance(int(headers.price))

    headers = Munch()
    headers.balance = str(channel.balance)
    headers.balance_signature = encode_hex(channel.balance_sig)
    headers.sender_address = channel.sender
    headers.open_block = str(channel.block)
    headers = HTTPHeaders.serialize(headers)

    response = requests.get(endpoint_url + '/resource', headers=headers)
    assert response.status_code == 200
    assert response.text.strip() == 'GET'

    assert_method(requests.post,
                  endpoint_url + '/resource',
                  headers, channel, 'POST',
                  expected_price=2)
    assert_method(requests.put,
                  endpoint_url + '/resource',
                  headers, channel, 'PUT',
                  expected_price=3)
    assert_method(requests.delete,
                  endpoint_url + '/resource',
                  headers, channel, 'DEL',
                  expected_price=4)



import pytest
import logging
import types
from microraiden import DefaultHTTPClient, Client
import microraiden.requests

log = logging.getLogger(__name__)


@pytest.fixture
def default_http_client(client: Client, use_tester: bool):
    # patch request_resource of this instance in order to advance blocks when doing requests
    def request_patched(self: DefaultHTTPClient, method: str, url: str, **kwargs):
        if use_tester:
            log.info('Mining new block.')
            self.client.context.web3.testing.mine(1)
        return DefaultHTTPClient._request_resource(self, method, url, **kwargs)

    http_client = DefaultHTTPClient(client, retry_interval=0.1)
    http_client._request_resource = types.MethodType(request_patched, http_client)
    yield http_client


@pytest.fixture
def init_microraiden_requests(default_http_client: DefaultHTTPClient):
    microraiden.requests.init(http_client=default_http_client)



import logging
import pytest
import gevent
import types

import rlp
from _pytest.monkeypatch import MonkeyPatch
from eth_utils import decode_hex
from ethereum.transactions import Transaction
import ethereum.tester
from populus.wait import Wait

from web3 import Web3, EthereumTesterProvider
from web3.contract import Contract
from web3.providers.rpc import RPCProvider

from microraiden.utils import (
    addr_from_sig,
    keccak256,
)
from microraiden.test.config import (
    CHANNEL_MANAGER_ADDRESS,
    FAUCET_ADDRESS,
    FAUCET_PRIVKEY,
    FAUCET_ALLOWANCE,
    INITIAL_TOKEN_SUPPLY
)
from microraiden.utils.contract import DEFAULT_TIMEOUT, DEFAULT_RETRY_INTERVAL
import microraiden.utils.contract


@pytest.fixture(scope='session')
def mine_sync_event():
    return gevent.event.Event()


@pytest.fixture(scope='session', autouse=True)
def disable_requests_loggin():
    logging.getLogger('requests').setLevel(logging.WARNING)
    logging.getLogger('urllib3').setLevel(logging.WARNING)


def deploy_token_contract(web3, deployer_address, token_abi, token_bytecode):
    Token = web3.eth.contract(abi=token_abi, bytecode=token_bytecode)
    txhash = Token.deploy(
        {'from': deployer_address}, args=[INITIAL_TOKEN_SUPPLY, "Raiden Network Token", "RDN", 18]
    )
    receipt = web3.eth.getTransactionReceipt(txhash)
    contract_address = receipt.contractAddress
    token = Token(contract_address)
    token.transact({'from': deployer_address}).transfer(FAUCET_ADDRESS, FAUCET_ALLOWANCE)
    assert web3.eth.getCode(contract_address) != '0x'
    return token


@pytest.fixture(scope='session')
def token_address(
        use_tester,
        web3,
        deployer_address,
        token_abi,
        token_bytecode,
        channel_manager_abi
):
    if use_tester:
        contract = deploy_token_contract(web3, deployer_address, token_abi, token_bytecode)
        return contract.address
    else:
        channel_manager = web3.eth.contract(
            abi=channel_manager_abi,
            address=CHANNEL_MANAGER_ADDRESS
        )
        return channel_manager.call().token()


def deploy_channel_manager_contract(
        web3,
        deployer_address,
        channel_manager_abi,
        channel_manager_bytecode,
        token_address
):
    ChannelManager = web3.eth.contract(abi=channel_manager_abi, bytecode=channel_manager_bytecode)
    txhash = ChannelManager.deploy({'from': deployer_address}, args=[token_address, 30])
    contract_address = web3.eth.getTransactionReceipt(txhash).contractAddress
    web3.testing.mine(1)

    return ChannelManager(contract_address)


@pytest.fixture(scope='session')
def channel_manager_address(
        use_tester,
        web3,
        deployer_address,
        channel_manager_abi,
        channel_manager_bytecode,
        token_address
):
    if use_tester:
        contract = deploy_channel_manager_contract(
            web3,
            deployer_address,
            channel_manager_abi,
            channel_manager_bytecode,
            token_address
        )
        return contract.address
    else:
        return CHANNEL_MANAGER_ADDRESS


@pytest.fixture(scope='session')
def web3(use_tester, mine_sync_event):
    if use_tester:
        provider = EthereumTesterProvider()
        web3 = Web3(provider)
        x = web3.testing.mine

        def mine_patched(self, count):
            x(count)
            mine_sync_event.set()
            gevent.sleep(0)  # switch context
            mine_sync_event.clear()

        web3.testing.mine = types.MethodType(
            mine_patched, web3.testing.mine
        )

        # Tester chain uses Transaction to send and validate transactions but does not support
        # EIP-155 yet. This patches the sender address recovery to handle EIP-155.
        sender_property_original = Transaction.sender.fget

        def sender_property_patched(self: Transaction):
            if self._sender:
                return self._sender

            if self.v and self.v >= 35:
                v = bytes([self.v])
                r = self.r.to_bytes(32, byteorder='big')
                s = self.s.to_bytes(32, byteorder='big')
                raw_tx = Transaction(
                    self.nonce, self.gasprice, self.startgas, self.to, self.value, self.data,
                    (self.v - 35) // 2, 0, 0
                )
                msg = keccak256(rlp.encode(raw_tx))
                self._sender = decode_hex(addr_from_sig(r + s + v, msg))
                return self._sender
            else:
                return sender_property_original(self)

        Transaction.sender = property(
            sender_property_patched,
            Transaction.sender.fset,
            Transaction.sender.fdel
        )

        # add faucet account to tester
        ethereum.tester.accounts.append(decode_hex(FAUCET_ADDRESS))
        ethereum.tester.keys.append(decode_hex(FAUCET_PRIVKEY))

        # make faucet rich
        web3.eth.sendTransaction({'to': FAUCET_ADDRESS, 'value': FAUCET_ALLOWANCE})

    else:
        rpc = RPCProvider('localhost', 8545)
        web3 = Web3(rpc)

    yield web3

    if use_tester:
        ethereum.tester.accounts.remove(decode_hex(FAUCET_ADDRESS))
        ethereum.tester.keys.remove(decode_hex(FAUCET_PRIVKEY))


@pytest.fixture
def patched_contract(use_tester: bool, monkeypatch: MonkeyPatch, web3: Web3):
    if use_tester:
        def patched_get_logs_raw(contract: Contract, filter_params):
            filter_ = contract.web3.eth.filter(filter_params)
            response = contract.web3.eth.getFilterLogs(filter_.filter_id)
            contract.web3.eth.uninstallFilter(filter_.filter_id)
            return response

        def patched_wait_for_event(wait: float):
            web3.testing.mine(1)

        def patched_wait_for_transaction(
            web3: Web3,
            tx_hash: str,
            timeout: int = DEFAULT_TIMEOUT,
            polling_interval: int = DEFAULT_RETRY_INTERVAL
        ):
            web3.testing.mine(1)
            tx_receipt = web3.eth.getTransactionReceipt(tx_hash)
            if tx_receipt is None:
                raise TimeoutError('Transaction {} was not mined.'.format(tx_hash))
            return tx_receipt

        monkeypatch.setattr(microraiden.utils.contract, '_get_logs_raw', patched_get_logs_raw)
        monkeypatch.setattr(microraiden.utils.contract, '_wait', patched_wait_for_event)
        monkeypatch.setattr(
            microraiden.utils.contract,
            'wait_for_transaction',
            patched_wait_for_transaction
        )


@pytest.fixture(scope='session')
def wait(web3, kovan_block_time):
    poll_interval = kovan_block_time / 2
    return Wait(web3, poll_interval=poll_interval)


@pytest.fixture(scope='session')
def wait_for_blocks(web3, kovan_block_time, use_tester):
    def wait_for_blocks(n):
        if use_tester:
            web3.testing.mine(n)
            gevent.sleep(0)
        else:
            target_block = web3.eth.blockNumber + n
            while web3.eth.blockNumber < target_block:
                gevent.sleep(kovan_block_time / 2)
    return wait_for_blocks


@pytest.fixture(scope='session')
def wait_for_transaction(wait):
    def wait_for_transaction(tx_hash):
        wait.for_receipt(tx_hash)
        gevent.sleep(0)
    return wait_for_transaction


@pytest.fixture
def revert_chain(web3: Web3, use_tester: bool):
    if use_tester:
        snapshot_id = web3.testing.snapshot()
        yield
        web3.testing.revert(snapshot_id)


@pytest.fixture(scope='session')
def token_contract(web3: Web3, token_address: str, token_abi):
    return web3.eth.contract(abi=token_abi, address=token_address)


@pytest.fixture(scope='session')
def channel_manager_contract(web3: Web3, channel_manager_address: str, channel_manager_abi):
    return web3.eth.contract(abi=channel_manager_abi, address=channel_manager_address)



#!/usr/bin/env python3

import logging

import click

from microraiden import DefaultHTTPClient
from microraiden.client.channel import Channel
from microraiden import Client

log = logging.getLogger(__name__)


@click.command()
@click.option('--api-endpoint', default='localhost', help='Address of the HTTP API server.')
@click.option('--api-port', default=5000)
@click.option('--datadir', help='Raiden data directory.')
@click.option('--rpc-endpoint', help='Address of the Ethereum RPC server.')
@click.option('--rpc-port', help='Ethereum RPC port.')
@click.option(
    '--key-path',
    required=True,
    help='Path to private key file.',
    type=click.Path(exists=True, dir_okay=False, resolve_path=True)
)
@click.option(
    '--key-password-path',
    default=None,
    help='Path to file containing password for private key.',
    type=click.Path(exists=True, dir_okay=False, resolve_path=True)
)
@click.option('--close-channels', default=False, is_flag=True, type=bool,
              help='Close all open channels before exiting.')
@click.option(
    '--channel-manager-address',
    help='Ethereum address of the channel manager contract.'
)
@click.option(
    '--contract-abi-path',
    help='Path to a file containing the ABIs for the token and channel manager.'
)
@click.option(
    '--token-address',
    help='Ethereum address of the token contract.'
)
def run(api_endpoint, api_port, **kwargs):
    exclude_kwargs = {'close_channels'}
    kwargs_client = {
        key: value for key, value in kwargs.items() if value and key not in exclude_kwargs
    }
    with Client(**kwargs_client) as client:
        m2mclient = DefaultHTTPClient(client, api_endpoint, api_port)
        resource = m2mclient.run('doggo.jpg')
        log.info('Response: {}'.format(resource))

        if kwargs['close_channels'] is True:
            for channel in client.channels:
                if channel.state == Channel.State.open:
                    channel.close_channel()


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    run()



import pytest  # noqa: F401

from microraiden import HTTPHeaders


def test_headers():
    TEST_SENDER_ADDR = '0xaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'
    TEST_RECEIVER_ADDR = '0xffffffffffffffffffffffffffffffffffffffff'
    headers_dict = {
        'Content-Type': 'application/json',
        'RDN-Cost': '5',
        'RDN-Receiver-Address': TEST_RECEIVER_ADDR
    }

    headers = HTTPHeaders.deserialize(headers_dict)
    assert headers.cost == '5'
    assert headers.receiver_address == TEST_RECEIVER_ADDR
    assert 'sender_address' not in headers

    headers_dict = HTTPHeaders.serialize(headers)
    assert len(headers_dict) == 2
    assert headers_dict['RDN-Cost'] == '5'
    assert headers_dict['RDN-Receiver-Address'] == TEST_RECEIVER_ADDR

    headers.sender_address = TEST_SENDER_ADDR
    assert headers.sender_address == TEST_SENDER_ADDR
    headers_dict = HTTPHeaders.serialize(headers)
    assert len(headers_dict) == 3
    assert headers_dict['RDN-Sender-Address'] == TEST_SENDER_ADDR



"""
This is dummy code showing how the minimal app could look like.
"""
import click
import re
from microraiden import Client
from microraiden import DefaultHTTPClient
import logging
import requests


@click.command()
@click.option(
    '--key-path',
    required=True,
    help='Path to private key file.',
    type=click.Path(exists=True, dir_okay=False, resolve_path=True)
)
@click.option(
    '--key-password-path',
    default=None,
    help='Path to file containing password for private key.',
    type=click.Path(exists=True, dir_okay=False, resolve_path=True)
)
@click.option('--resource', required=True, help='Get this resource.')
def run(key_path, key_password_path, resource):
    # create the client
    with Client(key_path=key_path, key_password_path=key_password_path) as client:
        m2mclient = DefaultHTTPClient(
            client,
            'localhost',
            5000
        )

        # Get the resource. If payment is required, client will attempt to create
        # a channel or will use existing one.
        status, headers, body = m2mclient.run(resource)
        if status == requests.codes.OK:
            if re.match('^text\/', headers['Content-Type']):
                logging.info("got the resource %s type=%s\n%s" % (
                    resource,
                    headers.get('Content-Type', '???'),
                    body))
            else:
                logging.info("got the resource %s type=%s" % (
                    resource,
                    headers.get('Content-Type', '???')))
        else:
            logging.error("error getting the resource. code=%d body=%s" %
                          (status, body.decode().strip()))


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    run()



# to fix requests's SSL infinite recursion bug, SSL must be
# monkeypatched first, and only then requests can be imported
# the bug occurs with Python3.6, requests=2.8.14 and gevent==1.2.2
from gevent import monkey
monkey.patch_ssl()
import requests # noqa

from .client import (
    Client,
    HTTPClient,
    DefaultHTTPClient
)

from .header import (
    HTTPHeaders
)

__all__ = [
    Client,
    HTTPClient,
    DefaultHTTPClient,
    HTTPHeaders
]



import logging
from werkzeug.datastructures import EnvironHeaders
from flask import Response, make_response, request
from microraiden import HTTPHeaders as header
from flask_restful.utils import unpack

from microraiden.channel_manager import (
    ChannelManager,
)
from microraiden.exceptions import (
    NoOpenChannel,
    InvalidBalanceProof,
    InvalidBalanceAmount,
    InsufficientConfirmations
)
import microraiden.config as config
from functools import wraps
from eth_utils import is_address

log = logging.getLogger(__name__)


class RequestData:
    def __init__(self, headers, cookies=None):
        """parse a flask request object and check if the data received are valid"""
        assert isinstance(headers, EnvironHeaders)
        self.check_headers(headers)
        if cookies:
            self.check_cookies(cookies)

    def check_cookies(self, cookies):
        if header.BALANCE_SIGNATURE in cookies:
            self.balance_signature = cookies.get(header.BALANCE_SIGNATURE)
        if header.OPEN_BLOCK in cookies:
            self.open_block_number = int(cookies.get(header.OPEN_BLOCK))
        if header.SENDER_BALANCE in cookies:
            self.balance = int(cookies.get(header.SENDER_BALANCE))
        if header.SENDER_ADDRESS in cookies:
            self.sender_address = cookies.get(header.SENDER_ADDRESS)

    def check_headers(self, headers):
        """Check if headers sent by the client are valid"""
        price = headers.get(header.PRICE, None)
        contract_address = headers.get(header.CONTRACT_ADDRESS, None)
        receiver_address = headers.get(header.RECEIVER_ADDRESS, None)
        sender_address = headers.get(header.SENDER_ADDRESS, None)
        payment = headers.get(header.PAYMENT, None)
        balance_signature = headers.get(header.BALANCE_SIGNATURE, None)
        open_block = headers.get(header.OPEN_BLOCK, None)
        balance = headers.get(header.BALANCE, None)
        if price:
            price = int(price)
        if open_block:
            open_block = int(open_block)
        if balance:
            balance = int(balance)
        if price and price < 0:
            raise ValueError("Price must be >= 0")
        if contract_address and not is_address(contract_address):
            raise ValueError("Invalid contract address")
        if receiver_address and not is_address(receiver_address):
            raise ValueError("Invalid receiver address")
        if sender_address and not is_address(sender_address):
            raise ValueError("Invalid sender address")
        if payment and not isinstance(payment, int):
            raise ValueError("Payment must be an integer")
        if open_block and open_block < 0:
            raise ValueError("Open block must be >= 0")
        if balance and balance < 0:
            raise ValueError("Balance must be >= 0")

        self.price = price
        self.contract_address = contract_address
        self.receiver_address = receiver_address
        self.payment = payment
        self.balance_signature = balance_signature
        self.sender_address = sender_address
        self.open_block_number = open_block
        self.balance = balance


class Paywall(object):
    def __init__(self,
                 channel_manager,
                 light_client_proxy=None
                 ):
        super().__init__()
        assert isinstance(channel_manager, ChannelManager)
        assert is_address(channel_manager.channel_manager_contract.address)
        assert is_address(channel_manager.receiver)
        self.contract_address = channel_manager.channel_manager_contract.address
        self.receiver_address = channel_manager.receiver
        self.channel_manager = channel_manager
        self.light_client_proxy = light_client_proxy

    def access(self, resource, method, *args, **kwargs):
        if self.channel_manager.node_online() is False:
            return "Ethereum node is not responding", 502
        if self.channel_manager.get_eth_balance() < config.PROXY_BALANCE_LIMIT:
            return "Channel manager ETH balance is below limit", 502
        try:
            data = RequestData(request.headers, request.cookies)
        except ValueError as e:
            return str(e), 409
        accepts_html = r'text/html' in request.headers.get('Accept', '')
        headers = {}

        price = resource.price()

        # payment required
        if price > 0:
            paywall, headers = self.paywall_check(price, data)
            if paywall and accepts_html is True:
                reply_data = resource.get_paywall(request.path)
                return self.reply_webui(reply_data, headers)
            elif paywall:
                return make_response('', 402, headers)

        # all ok, return actual content
        resp = method(request.path, *args, **kwargs)
        if isinstance(resp, Response):
            return resp
        else:
            data, code, resource_headers = unpack(resp)
            resource_headers.update(headers)
            return make_response(str(data), code, resource_headers)

    def paywall_check(self, price, data):
        """Check if the resource can be sent to the client.
        Returns (is_paywalled: Bool, http_headers: dict)
        """
        headers = self.generate_headers(price)
        if not data.balance_signature:
            return True, headers

        # try to get an existing channel
        try:
            channel = self.channel_manager.verify_balance_proof(
                data.sender_address, data.open_block_number,
                data.balance, data.balance_signature)
        except InsufficientConfirmations as e:
            headers.update({header.INSUF_CONFS: "1"})
            return True, headers
        except NoOpenChannel as e:
            headers.update({header.NONEXISTING_CHANNEL: "1"})
            return True, headers
        except (InvalidBalanceAmount, InvalidBalanceProof):
            headers.update({header.INVALID_PROOF: 1})
            return True, headers

        # set headers to reflect channel state
        assert channel.sender is not None
        assert channel.balance >= 0
        headers.update(
            {
                header.SENDER_ADDRESS: channel.sender,
                header.SENDER_BALANCE: channel.balance
            })
        if channel.last_signature is not None:
            headers.update({header.BALANCE_SIGNATURE: channel.last_signature})

        amount_sent = data.balance - channel.balance

        if amount_sent != 0 and amount_sent != price:
            headers[header.INVALID_AMOUNT] = 1
            #  if difference is 0, it will be handled by channel manager
            return True, headers

        # set the headers to reflect actual state of a channel
        try:
            self.channel_manager.register_payment(
                channel.sender,
                data.open_block_number,
                data.balance,
                data.balance_signature)
        except (InvalidBalanceAmount, InvalidBalanceProof):
            # balance sent to the proxy is less than in the previous proof
            return True, headers

        # all ok, return premium content
        return False, headers

    # when are these generated?
    def generate_headers(self, price: int):
        assert price > 0
        """Generate basic headers that are sent back for every request"""
        headers = {
            header.GATEWAY_PATH: config.API_PATH,
            header.RECEIVER_ADDRESS: self.receiver_address,
            header.CONTRACT_ADDRESS: self.contract_address,
            header.TOKEN_ADDRESS: self.channel_manager.get_token_address(),
            header.PRICE: price,
            'Content-Type': 'application/json'
        }
        return headers

    def reply_webui(self, reply_data='', headers: dict={}):
        headers.update({
            "Content-Type": "text/html",
        })
        reply = make_response(reply_data, 402, headers)
        for hdr in (header.GATEWAY_PATH,
                    header.CONTRACT_ADDRESS,
                    header.RECEIVER_ADDRESS,
                    header.PRICE,
                    header.NONEXISTING_CHANNEL,
                    header.TOKEN_ADDRESS):
            if hdr in headers:
                reply.set_cookie(hdr, str(headers[hdr]))
        return reply


def paywall_decorator(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        self = func.__self__  # get instance of the bound method
        return self.paywall.access(
            self,
            func,
            *args,
            **kwargs
        )
    return wrapper



import click
import os
import sys
#
# Flask restarts itself when a file changes, but this restart
#  does not have PYTHONPATH set properly if you start the
#  app with python -m microraiden.
#
from microraiden.utils import privkey_to_addr
import logging
import requests

log = logging.getLogger(__name__)


if __package__ is None:
    path = os.path.dirname(os.path.dirname(__file__))
    sys.path.insert(0, path)
    sys.path.insert(0, path + "/../")

from web3 import Web3, HTTPProvider
from microraiden.make_helpers import make_paywalled_proxy
from microraiden import utils, config
from microraiden.exceptions import StateFileLocked, InsecureStateFile, NetworkIdMismatch
from microraiden.proxy.paywalled_proxy import PaywalledProxy

pass_app = click.make_pass_decorator(PaywalledProxy)


@click.group()
@click.option(
    '--channel-manager-address',
    default=None,
    help='Ethereum address of the channel manager contract.'
)
@click.option(
    '--state-file',
    default=None,
    help='State file of the proxy'
)
@click.option(
    '--private-key',
    required=True,
    help='Path to private key file of the proxy',
    type=click.Path(exists=True, dir_okay=False, resolve_path=True)
)
@click.option(
    '--private-key-password-file',
    default=None,
    help='Path to file containing password for the JSON-encoded private key',
    type=click.Path(exists=True, dir_okay=False, resolve_path=True)
)
@click.option(
    '--ssl-cert',
    default=None,
    help='Cerfificate of the server (cert.pem or similar)'
)
@click.option(
    '--rpc-provider',
    default=config.WEB3_PROVIDER_DEFAULT,
    help='Address of the Ethereum RPC provider'
)
@click.option(
    '--ssl-key',
    default=None,
    help='SSL key of the server (key.pem or similar)'
)
@click.option(
    '--paywall-info',
    default=config.HTML_DIR,
    help='Directory where the paywall info is stored. '
         'The directory shoud contain a index.html file with the payment info/webapp. '
         'Content of the directory (js files, images..) is available on the "js/" endpoint.'
)
@click.pass_context
def main(
    ctx,
    channel_manager_address,
    ssl_key,
    ssl_cert,
    state_file,
    private_key,
    private_key_password_file,
    paywall_info,
    rpc_provider,
):
    private_key = utils.get_private_key(private_key, private_key_password_file)
    if private_key is None:
        sys.exit(1)

    receiver_address = privkey_to_addr(private_key)
    channel_manager_address = channel_manager_address or config.CHANNEL_MANAGER_ADDRESS

    if not state_file:
        state_file_name = "%s_%s.db" % (channel_manager_address[:10], receiver_address[:10])
        app_dir = click.get_app_dir('microraiden')
        if not os.path.exists(app_dir):
            os.makedirs(app_dir)
        state_file = os.path.join(app_dir, state_file_name)

    config.paywall_html_dir = paywall_info
    web3 = Web3(HTTPProvider(rpc_provider, request_kwargs={'timeout': 60}))
    try:
        app = make_paywalled_proxy(private_key, state_file,
                                   contract_address=channel_manager_address,
                                   web3=web3)
    except StateFileLocked as ex:
        log.fatal('Another uRaiden process is already running (%s)!' % str(ex))
        sys.exit(1)
    except InsecureStateFile as ex:
        msg = ('The permission bits of the state file (%s) are set incorrectly (others can '
               'read or write) or you are not the owner. For reasons of security, '
               'startup is aborted.' % state_file)
        log.fatal(msg)
        sys.exit(1)
    except NetworkIdMismatch as ex:
        log.fatal(str(ex))
        sys.exit(1)
    except requests.exceptions.ConnectionError as ex:
        log.fatal("Ethereum node refused connection: %s" % str(ex))
        sys.exit(1)
    ctx.obj = app



import logging
from enum import Enum

from eth_utils import decode_hex, is_same_address
from typing import Callable

from microraiden.client.context import Context
from microraiden.utils import (
    get_event_blocking,
    create_signed_contract_transaction,
    sign_balance_proof,
    verify_closing_sig,
    keccak256
)

log = logging.getLogger(__name__)


class Channel:
    class State(Enum):
        open = 1
        settling = 2
        closed = 3

    def __init__(
            self,
            core: Context,
            sender: str,
            receiver: str,
            block: int,
            deposit: int = 0,
            balance: int = 0,
            state: State = State.open,
            on_settle: Callable[['Channel'], None] = lambda channel: None
    ):
        self._balance = 0
        self._balance_sig = None

        self.core = core
        self.sender = sender.lower()
        self.receiver = receiver.lower()
        self.deposit = deposit
        self.block = block
        self.update_balance(balance)
        self.state = state
        self.on_settle = on_settle

        assert self.block is not None
        assert self._balance_sig

    @property
    def balance(self):
        return self._balance

    @property
    def key(self) -> bytes:
        return keccak256(self.sender, self.receiver, self.block)

    def update_balance(self, value):
        self._balance = value
        self._balance_sig = self.sign()

    @property
    def balance_sig(self):
        return self._balance_sig

    def sign(self):
        return sign_balance_proof(
            self.core.private_key,
            self.receiver,
            self.block,
            self.balance,
            self.core.channel_manager.address
        )

    def topup(self, deposit):
        """
        Attempts to increase the deposit in an existing channel. Block until confirmation.
        """
        if self.state != Channel.State.open:
            log.error('Channel must be open to be topped up.')
            return

        token_balance = self.core.token.call().balanceOf(self.core.address)
        if token_balance < deposit:
            log.error(
                'Insufficient tokens available for the specified topup ({}/{})'
                .format(token_balance, deposit)
            )
            return None

        log.info('Topping up channel to {} created at block #{} by {} tokens.'.format(
            self.receiver, self.block, deposit
        ))
        current_block = self.core.web3.eth.blockNumber

        data = decode_hex(self.receiver) + self.block.to_bytes(4, byteorder='big')
        tx = create_signed_contract_transaction(
            self.core.private_key,
            self.core.token,
            'transfer',
            [
                self.core.channel_manager.address,
                deposit,
                data
            ]
        )
        self.core.web3.eth.sendRawTransaction(tx)

        log.debug('Waiting for topup confirmation event...')
        event = get_event_blocking(
            self.core.channel_manager,
            'ChannelToppedUp',
            from_block=current_block + 1,
            argument_filters={
                '_sender': self.sender,
                '_receiver': self.receiver,
                '_open_block_number': self.block
            }
        )

        if event:
            log.debug('Successfully topped up channel in block {}.'.format(event['blockNumber']))
            self.deposit += deposit
            return event
        else:
            log.error('No event received.')
            return None

    def close(self, balance=None):
        """
        Attempts to request close on a channel. An explicit balance can be given to override the
        locally stored balance signature. Blocks until a confirmation event is received or timeout.
        """
        if self.state != Channel.State.open:
            log.error('Channel must be open to request a close.')
            return
        log.info('Requesting close of channel to {} created at block #{}.'.format(
            self.receiver, self.block
        ))
        current_block = self.core.web3.eth.blockNumber

        if balance is not None:
            self.update_balance(balance)

        tx = create_signed_contract_transaction(
            self.core.private_key,
            self.core.channel_manager,
            'uncooperativeClose',
            [
                self.receiver,
                self.block,
                self.balance,
                self.balance_sig
            ]
        )
        self.core.web3.eth.sendRawTransaction(tx)

        log.debug('Waiting for close confirmation event...')
        event = get_event_blocking(
            self.core.channel_manager,
            'ChannelCloseRequested',
            from_block=current_block + 1,
            argument_filters={
                '_sender': self.sender,
                '_receiver': self.receiver,
                '_open_block_number': self.block
            }
        )

        if event:
            log.debug('Successfully sent channel close request in block {}.'.format(
                event['blockNumber']
            ))
            self.state = Channel.State.settling
            return event
        else:
            log.error('No event received.')
            return None

    def close_cooperatively(self, closing_sig: bytes):
        """
        Attempts to close the channel immediately by providing a hash of the channel's balance
        proof signed by the receiver. This signature must correspond to the balance proof stored in
        the passed channel state.
        """
        if self.state == Channel.State.closed:
            log.error('Channel must not be closed already to be closed cooperatively.')
            return None
        log.info('Attempting to cooperatively close channel to {} created at block #{}.'.format(
            self.receiver, self.block
        ))
        current_block = self.core.web3.eth.blockNumber
        if not is_same_address(verify_closing_sig(self.balance_sig, closing_sig), self.receiver):
            log.error('Invalid closing signature.')
            return None

        tx = create_signed_contract_transaction(
            self.core.private_key,
            self.core.channel_manager,
            'cooperativeClose',
            [
                self.receiver,
                self.block,
                self.balance,
                self.balance_sig,
                closing_sig
            ]
        )
        self.core.web3.eth.sendRawTransaction(tx)

        log.debug('Waiting for settle confirmation event...')
        event = get_event_blocking(
            self.core.channel_manager,
            'ChannelSettled',
            from_block=current_block + 1,
            argument_filters={
                '_sender': self.sender,
                '_receiver': self.receiver,
                '_open_block_number': self.block
            }
        )

        if event:
            log.debug('Successfully closed channel in block {}.'.format(event['blockNumber']))
            self.state = Channel.State.closed
            return event
        else:
            log.error('No event received.')
            return None

    def settle(self):
        """
        Attempts to settle a channel that has passed its settlement period. If a channel cannot be
        settled yet, the call is ignored with a warning. Blocks until a confirmation event is
        received or timeout.
        """
        if self.state != Channel.State.settling:
            log.error('Channel must be in the settlement period to settle.')
            return None
        log.info('Attempting to settle channel to {} created at block #{}.'.format(
            self.receiver, self.block
        ))

        _, _, settle_block, _ = self.core.channel_manager.call().getChannelInfo(
            self.sender, self.receiver, self.block
        )

        current_block = self.core.web3.eth.blockNumber
        wait_remaining = settle_block - current_block
        if wait_remaining > 0:
            log.warning('{} more blocks until this channel can be settled. Aborting.'.format(
                wait_remaining
            ))
            return None

        tx = create_signed_contract_transaction(
            self.core.private_key,
            self.core.channel_manager,
            'settle',
            [
                self.receiver,
                self.block
            ]
        )
        self.core.web3.eth.sendRawTransaction(tx)

        log.debug('Waiting for settle confirmation event...')
        event = get_event_blocking(
            self.core.channel_manager,
            'ChannelSettled',
            from_block=current_block + 1,
            argument_filters={
                '_sender': self.sender,
                '_receiver': self.receiver,
                '_open_block_number': self.block
            }
        )

        if event:
            log.debug('Successfully settled channel in block {}.'.format(event['blockNumber']))
            self.state = Channel.State.closed
            self.on_settle(self)
            return event
        else:
            log.error('No event received.')
            return None

    def create_transfer(self, value):
        """
        Updates the given channel's balance and balance signature with the new value. The signature
        is returned and stored in the channel state.
        """
        assert value >= 0
        if value > self.deposit - self.balance:
            log.error(
                'Insufficient funds on channel. Needed: {}. Available: {}/{}.'
                .format(value, self.deposit - self.balance, self.deposit)
            )
            return None

        log.debug('Signing new transfer of value {} on channel to {} created at block #{}.'.format(
            value, self.receiver, self.block
        ))

        if self.state == Channel.State.closed:
            log.error('Channel must be open to create a transfer.')
            return None

        self.update_balance(self.balance + value)

        return self.balance_sig

    def is_valid(self) -> bool:
        return self.sign() == self.balance_sig and self.balance <= self.deposit

    def is_suitable(self, value: int):
        return self.deposit - self.balance >= value



import pytest  # noqa: F401
from coincurve import PublicKey
from eth_utils import encode_hex, decode_hex
from web3.contract import Contract

from microraiden.utils import (
    privkey_to_addr,
    keccak256_hex,
    sign,
    pubkey_to_addr,
    get_balance_message,
    sign_balance_proof,
    verify_balance_proof,
    eth_sign,
    keccak256,
    addr_from_sig,
    eth_verify,
    eth_sign_typed_data_message_eip,
    eth_sign_typed_data_eip,
    pack
)

SENDER_PRIVATE_KEY = '0xa0a0a0a0a0a0a0a0a0a0a0a0a0a0a0a0a0a0a0a0a0a0a0a0a0a0a0a0a0a0a0a0'
SENDER_ADDR = privkey_to_addr(SENDER_PRIVATE_KEY)
RECEIVER_PRIVATE_KEY = '0xb1b1b1b1b1b1b1b1b1b1b1b1b1b1b1b1b1b1b1b1b1b1b1b1b1b1b1b1b1b1b1b1'
RECEIVER_ADDR = privkey_to_addr(RECEIVER_PRIVATE_KEY)


def test_encode_hex():
    assert isinstance(encode_hex(b''), str)
    assert isinstance(decode_hex(''), bytes)


def test_pack():
    assert pack(False) == b'\x00'
    assert pack(True) == b'\x01'


def test_keccak256():
    addr1 = '0x1212121212121212121212121212121212121212'
    addr2 = '0x3434343434343434343434343434343434343434'

    ref = '0x3ac225168df54212a25c1c01fd35bebfea408fdac2e31ddd6f80a4bbf9a5f1cb'
    assert keccak256_hex('a') == ref

    ref = '0x0ef9d8f8804d174666011a394cab7901679a8944d24249fd148a6a36071151f8'
    assert keccak256_hex('0x0a') == ref

    ref = '0x1e17ec1733a759ef0d988fd0195d0e2792dc6181c8953130d14f9a261c6260bb'
    assert keccak256_hex(addr1, (15, 32)) == ref

    ref = '0x00424f9e17d5fecf44d798962d27cd1d44ef37d3968b05a5ff078ff696c14ea8'
    assert keccak256_hex(addr1, (17, 32), addr2) == ref

    ref = '0xc624b66cc0138b8fabc209247f72d758e1cf3343756d543badbf24212bed8c15'
    assert keccak256_hex((23, 256)) == ref

    ref = '0x66de8ffda797e3de9c05e8fc57b3bf0ec28a930d40b0d285d93c06501cf6a090'
    assert keccak256_hex(19) == ref

    ref = '0x7234c58e51ab4abdf62492ac6faf025ebff2afd4f861cebfa33d3e76667716a9'
    assert keccak256_hex(-5) == ref

    ref = '0x5fe7f977e71dba2ea1a68e21057beebb9be2ac30c6410aa38d4f3fbe41dcffd2'
    assert keccak256_hex(True) == ref

    ref = '0xbc36789e7a1e281436464229828f817d6612f7b477d66591ff96a9e064bcc98a'
    assert keccak256_hex(False) == ref


def test_sign():
    msg = b'32' * 16
    assert len(msg) == 32
    sig = sign(SENDER_PRIVATE_KEY, msg)
    pubkey = PublicKey.from_signature_and_message(sig, msg, hasher=None)
    pubkey = pubkey.format(compressed=False)
    assert len(sig) == 65
    assert pubkey_to_addr(pubkey) == SENDER_ADDR


def test_eth_sign():
    # Generated using https://www.myetherwallet.com/signmsg.html
    msg = 'is it wednesday, my dudes?'

    sig_expected = '0xcc7b4e6cde6ace1d99995661250e52388aae17ebd66dcf52e634a6dd51bf286a2' \
                   'e9757967b7baff8b549da7d8c3340701abf8560430b7a0bdf34f42b19bbf1861b'

    assert encode_hex(eth_sign(SENDER_PRIVATE_KEY, msg)) == sig_expected


def test_eth_sign_typed_data():
    pass


def test_eth_sign_typed_data_eip():
    # Test cases from the EIP:
    # https://github.com/0xProject/EIPs/blob/01dfc0f9a4122d8ad8817c503447cab8efa8a6c4/EIPS/eip-signTypedData.md#test-cases
    privkey = 'f2f48ee19680706196e2e339e5da3491186e0c4c5030670656b0e0164837257d'
    addr = '0x5409ed021d9299bf6814279a6a1411a7e866a631'
    assert addr == privkey_to_addr(privkey)

    typed_data = [('string', 'message', 'Hi, Alice!')]

    msg = eth_sign_typed_data_message_eip(typed_data)
    assert encode_hex(msg) == '0xe18794748cc6d73634d578f6a83f752bee11a0c9853d76bd0111d67a9b555a2c'

    sig = eth_sign_typed_data_eip(privkey, typed_data)
    sig_expected = '0x1a4ca93acf066a580f097690246e6c85d1deeb249194f6d3c2791f3aecb6adf8' \
                   '714ca4a0f12512ddd2a4f2393ea0c3b2c856279ba4929a5a34ae6859689428061b'
    assert encode_hex(sig) == sig_expected

    typed_data = [('uint', 'value', 42)]

    msg = eth_sign_typed_data_message_eip(typed_data)
    assert encode_hex(msg) == '0x6cb1c2645d841a0a3d142d1a2bdaa27015cc77f442e17037015b0350e468a957'

    sig = eth_sign_typed_data_eip(privkey, typed_data)
    sig_expected = '0x87c5b6a9f3a758babcc9140a96ae07957c6c9109af65bf139266cded52da49e6' \
                   '3df6af6f7daef588218e156bc83b95e0bfcfa8e72843cf4cf8c67c3ca11c3fd11b'
    assert encode_hex(sig) == sig_expected

    typed_data = [
        ('uint', 'value', 42),
        ('string', 'message', 'Hi, Alice!'),
        ('bool', 'removed', False)
    ]

    msg = eth_sign_typed_data_message_eip(typed_data)
    assert encode_hex(msg) == '0x36c3ed8591950e33dc4777bb455ab1a3e4223f84c42172a1ff2e200d5e25ee2e'

    sig = eth_sign_typed_data_eip(privkey, typed_data)
    sig_expected = '0x466a5a021225b681836e9951d2e603a37a605850ca26da0692ca532d4d1581d6' \
                   '031e2e3754fe31036e3a56d7e37fb6f598f7ab7a5cdd87aa04c9811c8b6209f31b'
    assert encode_hex(sig) == sig_expected


def test_get_balance_message(channel_manager_address: str):
    msg = get_balance_message(RECEIVER_ADDR, 310214, 14, channel_manager_address)
    assert encode_hex(msg) == '0x030600a234c173e696c945642673a00720f591bea0741589a3cb0cb09a898ce1'


def test_sign_balance_proof_contract(channel_manager_contract: Contract):
    sig = sign_balance_proof(
        SENDER_PRIVATE_KEY, RECEIVER_ADDR, 37, 15, channel_manager_contract.address
    )
    sender_recovered = channel_manager_contract.call().verifyBalanceProof(
        RECEIVER_ADDR, 37, 15, sig
    )
    assert sender_recovered == SENDER_ADDR


def test_verify_balance_proof(channel_manager_address: str):
    sig = sign_balance_proof(
        SENDER_PRIVATE_KEY, RECEIVER_ADDR, 315123, 8, channel_manager_address
    )
    assert verify_balance_proof(
        RECEIVER_ADDR, 315123, 8, sig, channel_manager_address
    ) == SENDER_ADDR


def test_sign_v0():
    msg = keccak256('hello v=0')
    sig = sign(SENDER_PRIVATE_KEY, msg)
    assert sig[-1] == 1
    assert addr_from_sig(sig, msg) == SENDER_ADDR


def test_eth_sign_v27():
    sig = eth_sign(SENDER_PRIVATE_KEY, 'hello v=27')
    assert sig[-1] == 27
    assert eth_verify(sig, 'hello v=27') == SENDER_ADDR


def test_verify_balance_proof_v0(channel_manager_address: str):
    sig = sign_balance_proof(
        SENDER_PRIVATE_KEY, RECEIVER_ADDR, 312524, 11, channel_manager_address
    )
    sig = sig[:-1] + b'\x00'
    assert verify_balance_proof(
        RECEIVER_ADDR, 312524, 11, sig, channel_manager_address
    ) == SENDER_ADDR


def test_verify_balance_proof_v27(channel_manager_address: str):
    # Should be default but test anyway.
    sig = sign_balance_proof(
        SENDER_PRIVATE_KEY, RECEIVER_ADDR, 312524, 11, channel_manager_address
    )
    sig = sig[:-1] + b'\x1b'
    assert verify_balance_proof(
        RECEIVER_ADDR, 312524, 11, sig, channel_manager_address
    ) == SENDER_ADDR



from .manager import ChannelManager
from .blockchain import Blockchain
from .state import ChannelManagerState
from .channel import Channel, ChannelState

__all__ = [
    ChannelManager,
    Blockchain,
    ChannelManagerState,
    Channel,
    ChannelState
]



import requests
import json
import time
import sys


class bcolors:
    HEADER = '\033[95m'
    BLUE = '\033[94m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    RED = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def nice_print(key, val, val_prev):
    if val > val_prev:
        color = bcolors.GREEN
    elif val < val_prev:
        color = bcolors.RED
    else:
        color = bcolors.ENDC
    print(bcolors.BOLD + key + "\t\t" + color + str(val) + bcolors.ENDC)


def do_request():
    json_response = {
        'balance_sum': 0,
        'open_channels': 0,
        'pending_channels': 0,
        'unique_senders': 0,
        'liquid_balance': 0,
        'deposit_sum': 0
    }
    json_prev = None
    connected = False
    while True:
        try:
            response = requests.get("http://localhost:5000/api/1/stats")
            json_response = json.loads(response.text)
            connected = True
        except requests.exceptions.ConnectionError:
            connected = False
            pass
        if json_prev is None:
            json_prev = json_response
        print('\033[H\033[J')
        print('connected' if connected else 'disconnected')
        nice_print("balance", json_response['balance_sum'], json_prev['balance_sum'])
        nice_print("open", json_response['open_channels'], json_prev['open_channels'])
        nice_print("pending", json_response['pending_channels'], json_prev['pending_channels'])
        nice_print("senders", json_response['unique_senders'], json_prev['unique_senders'])
        nice_print("liquid", json_response['liquid_balance'], json_prev['liquid_balance'])
        nice_print("deposit", json_response['deposit_sum'], json_prev['deposit_sum'])
        json_prev = json_response
        time.sleep(2)


def run():
    try:
        do_request()
    except KeyboardInterrupt:
        sys.exit(1)


if __name__ == '__main__':
    run()



import pytest
import time
import logging
import datetime

import gevent
from eth_utils import encode_hex

from microraiden import DefaultHTTPClient

log = logging.getLogger(__name__)


@pytest.mark.skip(reason="it takes too long")
def test_resource_request(doggo_proxy, default_http_client: DefaultHTTPClient):
    requests = 1000
    default_http_client.initial_deposit = lambda x: (requests + 1) * x

    # First transfer creates channel on-chain => exclude from profiling.
    response = default_http_client.run('doggo.jpg')
    assert response.decode().strip() == '"HI I AM A DOGGO"'

    t_start = time.time()
    for i in range(requests):
        log.debug('Transfer {}'.format(i))
        response = default_http_client.run('doggo.jpg')
        assert response.decode().strip() == '"HI I AM A DOGGO"'
    t_diff = time.time() - t_start

    log.info("{} requests in {} ({} rps)".format(
        requests, datetime.timedelta(seconds=t_diff), requests / t_diff)
    )


def test_receiver_validation(channel_manager, client, wait_for_blocks):
    n = 100
    # open channel
    channel = client.open_channel(channel_manager.state.receiver, n)
    wait_for_blocks(channel_manager.blockchain.n_confirmations)
    gevent.sleep(channel_manager.blockchain.poll_interval)
    assert (channel.sender, channel.block) in channel_manager.channels

    # prepare balance proofs
    t_start = time.time()
    balance_proofs = [encode_hex(channel.create_transfer(1)) for _ in range(n)]
    t_diff = time.time() - t_start
    log.info("%d balance proofs prepared in %s (%f / s)",
             n, datetime.timedelta(seconds=t_diff), n / t_diff)

    # validate
    t_start = time.time()
    for i, balance_proof in enumerate(balance_proofs):
        log.debug('Transfer {}'.format(i))
        sender, received = channel_manager.register_payment(
            channel.sender,
            channel.block,
            i + 1,
            balance_proof)
        assert sender == channel.sender
        assert received == 1

    t_diff = time.time() - t_start
    log.info("%d balance proofs verified in %s (%f / s)",
             n, datetime.timedelta(seconds=t_diff), n / t_diff)



import time
from enum import IntEnum


class ChannelState(IntEnum):
    OPEN = 0
    CLOSED = 1
    CLOSE_PENDING = 2
    UNDEFINED = 100


class Channel(object):
    """A channel between two parties."""

    def __init__(self, receiver, sender, deposit, open_block_number):
        self.receiver = receiver
        self.sender = sender  # sender address
        self.deposit = deposit  # deposit is maximum funds that can be used
        self.open_block_number = open_block_number

        self.balance = 0  # how much of the deposit has been spent
        self.state = ChannelState.UNDEFINED
        self.last_signature = None
        # if set, this is the absolute block_number the channel can be settled
        self.settle_timeout = -1
        self.ctime = time.time()  # channel creation time
        self.mtime = self.ctime
        self.confirmed = False

        self.unconfirmed_topups = {}  # txhash to added deposit

    @property
    def is_closed(self):
        return (self.state) in (ChannelState.CLOSED, ChannelState.CLOSE_PENDING)

    @is_closed.setter
    def is_closed(self, value):
        assert value is True
        self.state = ChannelState.CLOSED

    @property
    def unconfirmed_deposit(self):
        return self.deposit + sum(self.unconfirmed_topups.values())

    def to_dict(self):
        return self.__dict__

    @classmethod
    def from_dict(cls, state: dict):
        ret = cls(None, None, None, None)
        assert (set(state) - set(ret.__dict__)) == set()
        for k, v in state.items():
            if k in ret.__dict__.keys():
                setattr(ret, k, v)
        return ret



from typing import Dict
from munch import Munch


class HTTPHeaders(object):
    PRICE = 'RDN-Price'
    CONTRACT_ADDRESS = 'RDN-Contract-Address'
    RECEIVER_ADDRESS = 'RDN-Receiver-Address'
    TOKEN_ADDRESS = 'RDN-Token-Address'
    PAYMENT = 'RDN-Payment'
    BALANCE = 'RDN-Balance'
    BALANCE_SIGNATURE = 'RDN-Balance-Signature'
    SENDER_ADDRESS = 'RDN-Sender-Address'
    SENDER_BALANCE = 'RDN-Sender-Balance'
    GATEWAY_PATH = 'RDN-Gateway-Path'
    COST = 'RDN-Cost'
    OPEN_BLOCK = 'RDN-Open-Block'

# errors
    INSUF_FUNDS = 'RDN-Insufficient-Funds'
    INSUF_CONFS = 'RDN-Insufficient-Confirmations'
    NONEXISTING_CHANNEL = 'RDN-Nonexisting-Channel'
    INVALID_PROOF = 'RDN-Invalid-Balance-Proof'
    INVALID_AMOUNT = 'RDN-Invalid-Amount'

    DESERIALIZE_DICT = None
    SERIALIZE_DICT = None

    @classmethod
    def as_dict(cls):
        return dict((k.lower(), v) for k, v in cls.__dict__.items() if str(v).startswith('RDN'))

    @classmethod
    def deserialize(cls, headers: Dict[str, str] = None) -> Munch:
        if not cls.DESERIALIZE_DICT:
            cls.DESERIALIZE_DICT = {
                v.lower(): k.lower() for k, v in HTTPHeaders.__dict__.items()
                if str(v).startswith('RDN')
            }

        return Munch({
            cls.DESERIALIZE_DICT[key.lower()]: value for key, value in headers.items()
            if key.lower() in cls.DESERIALIZE_DICT
        })

    @classmethod
    def serialize(cls, headers: Munch) -> Dict[str, str]:
        if not cls.SERIALIZE_DICT:
            cls.SERIALIZE_DICT = {
                k.lower(): v for k, v in HTTPHeaders.__dict__.items() if str(v).startswith('RDN')
            }
        return {cls.SERIALIZE_DICT[k]: v for k, v in headers.items() if v}



from typing import List, Any, Union, Dict

import gevent
import rlp
from eth_utils import decode_hex, encode_hex
from ethereum.transactions import Transaction
from web3 import Web3
from web3.contract import Contract
from web3.formatters import input_filter_params_formatter, log_array_formatter
from web3.utils.events import get_event_data
from web3.utils.filters import construct_event_filter_params

from microraiden.config import GAS_PRICE, GAS_LIMIT
from microraiden.utils import privkey_to_addr, sign_transaction

DEFAULT_TIMEOUT = 60
DEFAULT_RETRY_INTERVAL = 3


def create_signed_transaction(
        private_key: str,
        web3: Web3,
        to: str,
        value: int=0,
        data=b'',
        nonce_offset: int = 0,
        gas_price: int = GAS_PRICE,
        gas_limit: int = GAS_LIMIT
) -> str:
    """
    Creates a signed on-chain transaction compliant with EIP155.
    """
    tx = create_transaction(
        web3=web3,
        from_=privkey_to_addr(private_key),
        to=to,
        value=value,
        data=data,
        nonce_offset=nonce_offset,
        gas_price=gas_price,
        gas_limit=gas_limit
    )
    sign_transaction(tx, private_key, web3.version.network)
    return encode_hex(rlp.encode(tx))


def create_transaction(
        web3: Web3,
        from_: str,
        to: str,
        data: bytes = b'',
        nonce_offset: int = 0,
        value: int = 0,
        gas_price: int = GAS_PRICE,
        gas_limit: int = GAS_LIMIT
) -> Transaction:
    nonce = web3.eth.getTransactionCount(from_, 'pending') + nonce_offset
    tx = Transaction(nonce, gas_price, gas_limit, to, value, data)
    tx.sender = decode_hex(from_)
    return tx


def create_signed_contract_transaction(
        private_key: str,
        contract: Contract,
        func_name: str,
        args: List[Any],
        value: int=0,
        nonce_offset: int = 0,
        gas_price: int = GAS_PRICE,
        gas_limit: int = GAS_LIMIT
) -> str:
    """
    Creates a signed on-chain contract transaction compliant with EIP155.
    """
    tx = create_contract_transaction(
        contract=contract,
        from_=privkey_to_addr(private_key),
        func_name=func_name,
        args=args,
        value=value,
        nonce_offset=nonce_offset,
        gas_price=gas_price,
        gas_limit=gas_limit
    )
    sign_transaction(tx, private_key, contract.web3.version.network)
    return encode_hex(rlp.encode(tx))


def create_contract_transaction(
        contract: Contract,
        from_: str,
        func_name: str,
        args: List[Any],
        value: int = 0,
        nonce_offset: int = 0,
        gas_price: int = GAS_PRICE,
        gas_limit: int = GAS_LIMIT
) -> Transaction:
    data = create_transaction_data(contract, func_name, args)
    return create_transaction(
        web3=contract.web3,
        from_=from_,
        to=contract.address,
        value=value,
        data=data,
        nonce_offset=nonce_offset,
        gas_price=gas_price,
        gas_limit=gas_limit
    )


def create_transaction_data(contract: Contract, func_name: str, args: List[Any]) -> bytes:
    data = contract._prepare_transaction(func_name, args)['data']
    return decode_hex(data)


def get_logs(
        contract: Contract,
        event_name: str,
        from_block: Union[int, str] = 0,
        to_block: Union[int, str] = 'pending',
        argument_filters: Dict[str, Any] = None
):
    event_abi = [
        abi_element for abi_element in contract.abi
        if abi_element['type'] == 'event' and abi_element['name'] == event_name
    ]
    assert len(event_abi) == 1, 'No event found matching name {}.'.format(event_name)
    event_abi = event_abi[0]

    if argument_filters is None:
        argument_filters = {}

    filter_params = input_filter_params_formatter(construct_event_filter_params(
        event_abi,
        argument_filters=argument_filters,
        address=contract.address,
        fromBlock=from_block,
        toBlock=to_block
    )[1])

    response = _get_logs_raw(contract, filter_params)

    logs = log_array_formatter(response)
    logs = [dict(log) for log in logs]
    for log in logs:
        log['args'] = get_event_data(event_abi, log)['args']
    return logs


def _get_logs_raw(contract: Contract, filter_params: Dict[str, Any]):
    """For easy patching."""
    return contract.web3._requestManager.request_blocking('eth_getLogs', [filter_params])


def get_event_blocking(
        contract: Contract,
        event_name: str,
        from_block: Union[int, str] = 0,
        to_block: Union[int, str] = 'pending',
        argument_filters: Dict[str, Any]=None,
        condition=None,
        wait=DEFAULT_RETRY_INTERVAL,
        timeout=DEFAULT_TIMEOUT
) -> Union[Dict[str, Any], None]:
    for i in range(0, timeout + wait, wait):
        logs = get_logs(
            contract,
            event_name,
            from_block=from_block,
            to_block=to_block,
            argument_filters=argument_filters
        )
        matching_logs = [event for event in logs if not condition or condition(event)]
        if matching_logs:
            return matching_logs[0]
        elif i < timeout:
            _wait(wait)

    return None


def _wait(duration: float):
    """For easy patching."""
    gevent.sleep(duration)


def wait_for_transaction(
        web3: Web3,
        tx_hash: str,
        timeout: int = DEFAULT_TIMEOUT,
        polling_interval: int = DEFAULT_RETRY_INTERVAL
):
    for waited in range(0, timeout + polling_interval, polling_interval):
        tx_receipt = web3.eth.getTransactionReceipt(tx_hash)
        if tx_receipt is not None:
            return tx_receipt
        if waited < timeout:
            _wait(polling_interval)
    raise TimeoutError('Transaction {} was not mined.'.format(tx_hash))



from web3 import Web3

from microraiden.config import CONTRACT_METADATA, TOKEN_ABI_NAME, CHANNEL_MANAGER_ABI_NAME
from microraiden.utils import privkey_to_addr


class Context(object):
    def __init__(
            self,
            private_key: str,
            web3: Web3,
            channel_manager_address: str
    ):
        self.private_key = private_key
        self.address = privkey_to_addr(private_key)
        self.web3 = web3

        self.channel_manager = web3.eth.contract(
            address=channel_manager_address,
            abi=CONTRACT_METADATA[CHANNEL_MANAGER_ABI_NAME]['abi']
        )

        token_address = self.channel_manager.call().token()
        self.token = web3.eth.contract(
            address=token_address,
            abi=CONTRACT_METADATA[TOKEN_ABI_NAME]['abi']
        )



from typing import List, Tuple, Any

from coincurve import PrivateKey, PublicKey
from eth_utils import encode_hex, decode_hex, remove_0x_prefix, keccak, is_0x_prefixed
from ethereum.transactions import Transaction
import rlp


Type = str
Name = str
TypedData = Tuple[Type, Name, Any]


def generate_privkey() -> bytes:
    return encode_hex(PrivateKey().secret)


def pubkey_to_addr(pubkey) -> str:
    if isinstance(pubkey, PublicKey):
        pubkey = pubkey.format(compressed=False)
    assert isinstance(pubkey, bytes)
    return encode_hex(keccak256(pubkey[1:])[-20:])


def privkey_to_addr(privkey: str) -> str:
    return pubkey_to_addr(PrivateKey.from_hex(remove_0x_prefix(privkey)).public_key)


def addr_from_sig(sig: bytes, msg: bytes):
    assert len(sig) == 65
    # Support Ethereum's EC v value of 27 and EIP 155 values of > 35.
    if sig[-1] >= 35:
        network_id = (sig[-1] - 35) // 2
        sig = sig[:-1] + bytes([sig[-1] - 35 - 2 * network_id])
    elif sig[-1] >= 27:
        sig = sig[:-1] + bytes([sig[-1] - 27])

    receiver_pubkey = PublicKey.from_signature_and_message(sig, msg, hasher=None)
    return pubkey_to_addr(receiver_pubkey)


def pack(*args) -> bytes:
    """
    Simulates Solidity's keccak256 packing. Integers can be passed as tuples where the second tuple
    element specifies the variable's size in bits, e.g.:
    keccak256((5, 32))
    would be equivalent to Solidity's
    keccak256(uint32(5))
    Default size is 256.
    """
    def format_int(value, size):
        assert isinstance(value, int)
        assert isinstance(size, int)
        if value >= 0:
            return decode_hex('{:x}'.format(value).zfill(size // 4))
        else:
            return decode_hex('{:x}'.format((1 << size) + value))

    msg = b''
    for arg in args:
        assert arg is not None
        if isinstance(arg, bytes):
            msg += arg
        elif isinstance(arg, str):
            if is_0x_prefixed(arg):
                msg += decode_hex(arg)
            else:
                msg += arg.encode()
        elif isinstance(arg, bool):
            msg += format_int(int(arg), 8)
        elif isinstance(arg, int):
            msg += format_int(arg, 256)
        elif isinstance(arg, tuple):
            msg += format_int(arg[0], arg[1])
        else:
            raise ValueError('Unsupported type: {}.'.format(type(arg)))

    return msg


def keccak256(*args) -> bytes:
    return keccak(pack(*args))


def keccak256_hex(*args) -> bytes:
    return encode_hex(keccak256(*args))


def sign(privkey: str, msg: bytes, v=0) -> bytes:
    assert isinstance(msg, bytes)
    assert isinstance(privkey, str)

    pk = PrivateKey.from_hex(remove_0x_prefix(privkey))
    assert len(msg) == 32

    sig = pk.sign_recoverable(msg, hasher=None)
    assert len(sig) == 65

    sig = sig[:-1] + bytes([sig[-1] + v])

    return sig


def sign_transaction(tx: Transaction, privkey: str, network_id: int):
    # Implementing EIP 155.
    tx.v = network_id
    sig = sign(privkey, keccak256(rlp.encode(tx)), v=35 + 2 * network_id)
    v, r, s = sig[-1], sig[0:32], sig[32:-1]
    tx.v = v
    tx.r = int.from_bytes(r, byteorder='big')
    tx.s = int.from_bytes(s, byteorder='big')


def eth_message_hash(msg: str) -> bytes:
    msg = '\x19Ethereum Signed Message:\n' + str(len(msg)) + msg
    return keccak256(msg)


def eth_sign(privkey: str, msg: str) -> bytes:
    assert isinstance(msg, str)
    sig = sign(privkey, eth_message_hash(msg), v=27)
    return sig


def eth_verify(sig: bytes, msg: str) -> str:
    return addr_from_sig(sig, eth_message_hash(msg))


def eth_sign_typed_data_message(typed_data: List[TypedData]) -> bytes:
    typed_data = [('{} {}'.format(type_, name), data) for type_, name, data in typed_data]
    schema, data = [list(zipped) for zipped in zip(*typed_data)]

    return keccak256(keccak256(*schema), keccak256(*data))


def eth_sign_typed_data(privkey: str, typed_data: List[TypedData]) -> bytes:
    msg = eth_sign_typed_data_message(typed_data)
    return sign(privkey, msg, v=27)


def eth_sign_typed_data_message_eip(typed_data: List[TypedData]) -> bytes:
    typed_data = [('{} {}'.format(type_, name), data) for type_, name, data in typed_data]
    schema, data = [list(zipped) for zipped in zip(*typed_data)]

    return keccak256(keccak256(*schema), *data)


def eth_sign_typed_data_eip(privkey: str, typed_data: List[TypedData]) -> bytes:
    msg = eth_sign_typed_data_message_eip(typed_data)
    return sign(privkey, msg, v=27)


def get_balance_message(
        receiver: str, open_block_number: int, balance: int, contract_address: str
) -> bytes:
    return eth_sign_typed_data_message([
        ('address', 'receiver', receiver),
        ('uint32', 'block_created', (open_block_number, 32)),
        ('uint192', 'balance', (balance, 192)),
        ('address', 'contract', contract_address)
    ])


def sign_balance_proof(
        privkey: str, receiver: str, open_block_number: int, balance: int, contract_address: str
) -> bytes:
    msg = get_balance_message(receiver, open_block_number, balance, contract_address)
    return sign(privkey, msg, v=27)


def verify_balance_proof(
        receiver: str,
        open_block_number: int,
        balance: int,
        balance_sig: bytes,
        contract_address: str
) -> str:
    msg = get_balance_message(receiver, open_block_number, balance, contract_address)
    return addr_from_sig(balance_sig, msg)


def sign_close(privkey: str, balance_sig: bytes) -> bytes:
    return sign(privkey, keccak256(balance_sig))


def verify_closing_sig(balance_sig: bytes, closing_sig: bytes) -> str:
    msg = keccak256(balance_sig)
    return addr_from_sig(closing_sig, msg)



from tkinter import ttk
import tkinter
import logging
import gevent

import click
import os

from microraiden import Client, DefaultHTTPClient
from microraiden.utils import privkey_to_addr
from microraiden.config import CHANNEL_MANAGER_ADDRESS, TKN_DECIMALS
from microraiden.proxy import PaywalledProxy
from microraiden.proxy.resources import PaywalledProxyUrl
from microraiden.make_helpers import make_paywalled_proxy

log = logging.getLogger(__name__)


def start_proxy(receiver_privkey: str) -> PaywalledProxy:
    state_file_name = '{}_{}.json'.format(
        CHANNEL_MANAGER_ADDRESS, privkey_to_addr(receiver_privkey)
    )
    app_dir = click.get_app_dir('microraiden')
    if not os.path.exists(app_dir):
        os.makedirs(app_dir)

    app = make_paywalled_proxy(receiver_privkey, os.path.join(app_dir, state_file_name))
    app.run()
    return app


class ETHTickerProxy:
    def __init__(self, privkey: str, proxy: PaywalledProxy = None) -> None:
        if proxy:
            self.app = proxy
        else:
            self.app = start_proxy(privkey)
        cfg = {'resource_class_kwargs': {
               'domain': 'http://api.bitfinex.com/v1/pubticker/'}
               }
        self.app.add_paywalled_resource(
            PaywalledProxyUrl,
            '/<string:ticker>',
            1 * TKN_DECIMALS,
            **cfg
        )

    def stop(self):
        self.app.stop()


class ETHTickerClient(ttk.Frame):
    def __init__(
            self,
            sender_privkey: str,
            httpclient: DefaultHTTPClient = None,
            poll_interval: float = 5
    ) -> None:
        self.poll_interval = poll_interval

        self.root = tkinter.Tk()
        ttk.Frame.__init__(self, self.root)
        self.root.title('µRaiden ETH Ticker')
        self.root.protocol('WM_DELETE_WINDOW', self.close)
        self.pack()
        self.pricevar = tkinter.StringVar(value='0.00 USD')
        ttk.Label(self, textvariable=self.pricevar, font=('Helvetica', '72')).pack()

        if httpclient:
            self.httpclient = httpclient
            self.client = httpclient.client
        else:
            self.client = Client(sender_privkey)
            self.httpclient = DefaultHTTPClient(
                self.client,
                initial_deposit=lambda x: 10 * x,
                topup_deposit=lambda x: 5 * x
            )

        self.active_query = False
        self.running = False

    def run(self):
        self.running = True
        self.root.after(0, self.query_price)
        self.root.mainloop()

    def query_price(self):
        if not self.running:
            return
        self.active_query = True

        response = self.httpclient.get('http://localhost:5000/ETHUSD')
        if response:
            price = float(response.json()['last_price'])
            log.info('New price received: {:.2f} USD'.format(price))
            self.pricevar.set('{:.2f} USD'.format(price))
        else:
            log.warning('No response.')

        if self.running:
            self.root.after(int(self.poll_interval * 1000), self.query_price)
        self.active_query = False

    def close(self):
        log.info('Shutting down gracefully.')
        self.running = False
        self.root.destroy()
        # Sloppy handling of thread joining but works for this small demo.
        while self.active_query:
            gevent.sleep(1)

        self.httpclient.close_active_channel('http://localhost:5000')


@click.command()
@click.option('--start-proxy/--no-proxy', default=False)
def main(start_proxy):
    proxy = None
    ticker = None
    receiver_privkey = 'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'
    sender_privkey = 'bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb'
    try:
        if start_proxy:
            proxy = ETHTickerProxy(receiver_privkey)
        ticker = ETHTickerClient(sender_privkey)
        ticker.run()
    except KeyboardInterrupt:
        if ticker:
            ticker.close()
        if proxy is not None:
            proxy.stop()


if __name__ == '__main__':
    from gevent import monkey
    monkey.patch_all()
    logging.basicConfig(level=logging.INFO)
    main()



from microraiden.proxy.resources.expensive import Expensive
from bs4 import BeautifulSoup
import os
import requests
import logging
from flask import Response, stream_with_context, request

from microraiden.config import MICRORAIDEN_DIR

log = logging.getLogger(__name__)


class PaywalledProxyUrl(Expensive):
    def __init__(self, domain=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.paywall_html = self.extract_paywall_body(
            os.path.join(MICRORAIDEN_DIR, 'microraiden/webui/index.html')
        )
        self.domain = domain

    def extract_paywall_body(self, path):
        # extract body of the paywall page and transform it into a div we'll be
        #  using later
        with open(path) as fp:
            soup = BeautifulSoup(fp, 'html.parser')
        b = soup.body
        b['id'] = "overlay"
        b.name = "div"
        return b

    def get(self, url, *args, **kwargs):
        req = requests.get(self.domain + url, stream=True, params=request.args)
        return Response(stream_with_context(req.iter_content()),
                        content_type=req.headers['content-type'])

    def get_paywall(self, url: str):
        data = self.get(url)

#  <link rel="stylesheet" type="text/css" href="/js/styles.css">

        soup = BeautifulSoup(data.data.decode(), 'html.parser')
        # generate js paths that are required
        js_paths = [
            "//code.jquery.com/jquery-3.2.1.js",
            "//cdnjs.cloudflare.com/ajax/libs/js-cookie/2.1.4/js.cookie.min.js",
            "//maxcdn.bootstrapcdn.com/bootstrap/3.3.7/js/bootstrap.min.js",
            "/js/web3.js",
            "/js/microraiden.js"]
        for src in js_paths:
            js_tag = soup.new_tag('script', type="text/javascript", src=src)
            soup.head.append(js_tag)
        # generate css
        bs_tag = soup.new_tag('link', rel="stylesheet",
                              type="text/css", href="/js/dark-bootstrap.min.css")
        css_tag = soup.new_tag('link', rel="stylesheet", type="text/css", href="/js/styles.css")
        soup.head.insert(0, bs_tag)
        soup.head.insert(0, css_tag)

        # inject div that generates the paywall
        soup.body.insert(0, self.paywall_html)

        return str(soup)



import pytest
from web3 import Web3

from microraiden import Client
from microraiden.utils import privkey_to_addr
from microraiden.test.utils.client import close_all_channels_cooperatively


@pytest.fixture
def client_address(client_privkey):
    return privkey_to_addr(client_privkey)


@pytest.fixture
def rpc_endpoint():
    return 'localhost'


@pytest.fixture
def rpc_port():
    return 8545


@pytest.fixture
def datadir(tmpdir):
    return tmpdir.strpath + "client"


@pytest.fixture
def client(
        sender_privkey: str,
        channel_manager_address: str,
        web3: Web3,
        clean_channels: bool,
        receiver_privkey: str,
        patched_contract,
        revert_chain
):
    client = Client(
        private_key=sender_privkey,
        channel_manager_address=channel_manager_address,
        web3=web3
    )
    if clean_channels:
        close_all_channels_cooperatively(client, receiver_privkey, balance=0)

    yield client

    if clean_channels:
        close_all_channels_cooperatively(client, receiver_privkey, balance=0)



import pytest
from eth_utils import encode_hex, remove_0x_prefix
from ethereum.tester import keys

import os
import json
from microraiden.utils import privkey_to_addr
from microraiden.config import CONTRACTS_ABI_JSON, CHANNEL_MANAGER_ABI_NAME, TOKEN_ABI_NAME


@pytest.fixture
def proxy_ssl():
    return False


@pytest.fixture
def test_dir(request):
    return request.fspath.dirname


@pytest.fixture
def proxy_ssl_certs(test_dir):
    return os.path.join(test_dir + '/data/key.pem'), os.path.join(test_dir + '/data/cert.pem')


@pytest.fixture(scope='session')
def use_tester(request):
    return request.config.getoption('use_tester')


@pytest.fixture(scope='session')
def clean_channels(request):
    return request.config.getoption('clean_channels')


@pytest.fixture
def api_endpoint():
    """address of a paywall proxy"""
    return 'localhost'


@pytest.fixture
def api_endpoint_port():
    """port the paywall proxy listens on"""
    return 5000


@pytest.fixture
def api_endpoint_address(api_endpoint, api_endpoint_port):
    return api_endpoint + ":" + str(api_endpoint_port)


@pytest.fixture
def init_contract_address():
    return "0x" + "a" * 40


@pytest.fixture(scope='session')
def deployer_privkey():
    return remove_0x_prefix(encode_hex(keys[3]))


@pytest.fixture(scope='session')
def deployer_address(deployer_privkey):
    return privkey_to_addr(deployer_privkey)


@pytest.fixture(scope='session')
def contract_abi_path():
    return os.path.join(os.path.dirname(os.path.dirname(__file__)), "../" + CONTRACTS_ABI_JSON)


@pytest.fixture(scope='session')
def contract_abis(contract_abi_path):
    with open(contract_abi_path) as abi_file:
        return json.load(abi_file)


@pytest.fixture(scope='session')
def channel_manager_abi(contract_abis):
    return contract_abis[CHANNEL_MANAGER_ABI_NAME]['abi']


@pytest.fixture(scope='session')
def channel_manager_bytecode(contract_abis):
    return contract_abis[CHANNEL_MANAGER_ABI_NAME]['bytecode']


@pytest.fixture(scope='session')
def token_abi(contract_abis):
    return contract_abis[TOKEN_ABI_NAME]['abi']


@pytest.fixture(scope='session')
def token_bytecode(contract_abis):
    return contract_abis[TOKEN_ABI_NAME]['bytecode']


@pytest.fixture(scope='session')
def kovan_block_time():
    return 4


@pytest.fixture
def state_db_path(tmpdir):
    return ':memory:'
#    db = tmpdir.join("state.db")
#    return db.strpath



from eth_utils import denoms
from web3 import Web3
from web3.utils.empty import empty as web3_empty

from microraiden.test.config import FAUCET_PRIVKEY
from microraiden.utils import create_signed_transaction, wait_for_transaction


def test_create_signed_transaction():
    # Mock to simulate the example from EIP 155.
    # https://github.com/ethereum/EIPs/blob/master/EIPS/eip-155.md#list-of-chain-ids
    class Web3Mock:
        class VersionMock:
            network = 1

        class EthMock:
            class ContractMock:
                def _prepare_transaction(self, *args, **kwargs):
                    return {'data': b''}

            defaultAccount = web3_empty

            def contract(self, *args, **kwargs):
                return self.ContractMock()

            def getTransactionCount(self, *args, **kwargs):
                return 9

            def getCode(self, *args, **kwargs):
                """Need to implement this to fake contract existence check"""
                return '0x123456789abcdef'

        version = VersionMock()
        eth = EthMock()

    # TODO: replace with proper mock
    web3 = Web3Mock()
    private_key = '0x4646464646464646464646464646464646464646464646464646464646464646'
    address = '0x3535353535353535353535353535353535353535'

    tx = create_signed_transaction(
        private_key,
        web3,
        to=address,
        value=10**18,
        gas_limit=21000,
        gas_price=20 * denoms.gwei
    )
    tx_expected = \
        '0xf86c098504a817c800825208943535353535353535353535353535353535353535880de0' \
        'b6b3a76400008025a028ef61340bd939bc2195fe537567866003e1a15d3c71ff63e1590620' \
        'aa636276a067cbe9d8997f761aecb703304b3800ccf555c9f3dc64214b297fb1966a3b6d83'
    assert tx == tx_expected


def test_wait_for_transaction(web3: Web3, patched_contract, receiver_address: str):
    tx = create_signed_transaction(
        FAUCET_PRIVKEY,
        web3,
        to=receiver_address
    )
    tx_hash = web3.eth.sendRawTransaction(tx)
    tx_receipt = wait_for_transaction(web3, tx_hash)
    assert tx_receipt



from flask_restful import Resource, reqparse
import json
from collections import defaultdict

from microraiden.utils import sign_close
from microraiden.proxy.resources.login import auth
from eth_utils import encode_hex

from microraiden.channel_manager import Channel, ChannelManager
from microraiden.exceptions import NoOpenChannel


class ChannelManagementRoot(Resource):
    @staticmethod
    def get():
        return "OK"


class ChannelManagementStats(Resource):
    def __init__(self, channel_manager: ChannelManager):
        super(ChannelManagementStats, self).__init__()
        self.channel_manager = channel_manager

    def get(self):
        deposit_sum = sum([c.deposit for c in self.channel_manager.channels.values()])
        unique_senders = {}
        open_channels = []
        pending_channels = []
        for k, v in self.channel_manager.channels.items():
            unique_senders[k[0]] = 1
            if v.is_closed is True:
                pending_channels.append(v)
            else:
                open_channels.append(v)
        contract_address = self.channel_manager.channel_manager_contract.address
        return {'balance_sum': self.channel_manager.get_locked_balance(),
                'deposit_sum': deposit_sum,
                'open_channels': len(open_channels),
                'pending_channels': len(pending_channels),
                'unique_senders': len(unique_senders),
                'liquid_balance': self.channel_manager.get_liquid_balance(),
                'token_address': self.channel_manager.token_contract.address,
                'contract_address': contract_address,
                'receiver_address': self.channel_manager.receiver,
                'manager_abi': self.channel_manager.channel_manager_contract.abi,
                'token_abi': self.channel_manager.token_contract.abi
                }


class ChannelManagementListChannels(Resource):
    def __init__(self, channel_manager: ChannelManager):
        super(ChannelManagementListChannels, self).__init__()
        self.channel_manager = channel_manager

    def get_all_channels(self, channel_status='all', condition=lambda k, v: True):
        return [
            {'sender_address': k[0],
             'open_block': k[1],
             'state': self.get_channel_status(v),
             'deposit': v.deposit,
             'balance': v.balance} for k, v in
            self.channel_manager.channels.items()
            if (condition(k, v))]

    def get_channel_filter(self, channel_status='all'):
        if channel_status == 'open' or channel_status == 'opened':
            return lambda c: c.is_closed is False
        elif channel_status == 'closed':
            return lambda c: c.is_closed is True
        else:
            return lambda c: True

    def get_channel_status(self, channel: Channel):
        if channel.is_closed is True:
            return "closed"
        elif channel.is_closed is False:
            return "open"
        else:
            return "unknown"

    def get(self, sender_address=None):
        parser = reqparse.RequestParser()
        parser.add_argument('status', help='filter channels by a status', default='open',
                            choices=('closed', 'opened', 'open', 'all'))
        args = parser.parse_args()
        channel_filter = self.get_channel_filter(args['status'])

        # if sender exists, return all open blocks
        if sender_address is not None:
            ret = self.get_all_channels(
                condition=lambda k, v:
                (k[0] == sender_address.lower() and
                 channel_filter(v))
            )

        # if sender is not specified, return all open channels
        else:
            channels = self.get_all_channels(condition=lambda k, v:
                                             channel_filter(v))
            joined_channels = defaultdict(list)
            for c in channels:
                joined_channels[c['sender_address']].append(c['open_block'])
            ret = [
                {'sender_address': k,
                 'blocks': v
                 } for k, v in joined_channels.items()
            ]

        return json.dumps(ret), 200

    def delete(self, sender_address):
        parser = reqparse.RequestParser()
        parser.add_argument('open_block', type=int, help='block the channel was opened')
        parser.add_argument('signature', help='last balance proof signature')
        args = parser.parse_args()
        if args.signature is None:
            return "Bad signature format", 400
        if args.block is None:
            return "No opening block specified", 400
        channel = self.channel_manager.channels[sender_address, args.block]
        if channel.last_signature != args.signature:
            return "Invalid or outdated balance signature", 400
        ret = sign_close(self.channel_manager.private_key, args.signature)
        return ret, 200


class ChannelManagementChannelInfo(Resource):
    def __init__(self, channel_manager):
        super(ChannelManagementChannelInfo, self).__init__()
        self.channel_manager = channel_manager

    def get(self, sender_address, opening_block):
        try:
            key = (sender_address.lower(), opening_block)
            sender_channel = self.channel_manager.channels[key]
        except KeyError:
            return "Sender address not found", 404

        return sender_channel.to_dict(), 200

    def delete(self, sender_address, opening_block):
        parser = reqparse.RequestParser()
        parser.add_argument('balance', type=int, help='last balance proof balance')
        args = parser.parse_args()
        if args.balance is None:
            return "Bad balance format", 400

        try:
            close_signature = self.channel_manager.sign_close(
                sender_address,
                opening_block,
                args.balance)
        except NoOpenChannel as e:
            return str(e), 400
        except KeyError:
            return "Channel not found", 404
        ret = {'close_signature': encode_hex(close_signature)}

        return ret, 200


class ChannelManagementAdminChannels(Resource):
    def __init__(self, channel_manager):
        super(ChannelManagementAdminChannels, self).__init__()
        self.channel_manager = channel_manager

    @auth.login_required
    def delete(self, sender_address, opening_block):
        self.channel_manager.force_close_channel(sender_address, opening_block)
        return "force closed (%s, %d)" % (sender_address, opening_block), 200


class ChannelManagementAdmin(Resource):
    def __init__(self, channel_manager):
        super(ChannelManagementAdmin, self).__init__()
        self.channel_manager = channel_manager

    @auth.login_required
    def get(self):
        return "NOTHING TO SEE HERE, GO AWAY", 200



from typing import Callable

from requests import Response
from web3 import Web3

from microraiden import DefaultHTTPClient, Client, HTTPClient
from microraiden.config import CHANNEL_MANAGER_ADDRESS

_http_client = None


def init(
    http_client: HTTPClient = None,
    client: Client = None,
    privkey: str = None,
    key_path: str = None,
    key_password_path: str = None,
    channel_manager_address: str = CHANNEL_MANAGER_ADDRESS,
    web3: Web3 = None,
    retry_interval: float = 5,
    initial_deposit: Callable[[int], int] = lambda price: 10 * price,
    topup_deposit: Callable[[int], int] = lambda price: 5 * price
):
    """
    TODO: document which of these arguments are actually needed in different use cases.
    """
    global _http_client

    if http_client is None:
        if client is None:
            client = Client(
                private_key=privkey,
                key_path=key_path,
                key_password_path=key_password_path,
                channel_manager_address=channel_manager_address,
                web3=web3
            )
        _http_client = DefaultHTTPClient(
            client=client,
            retry_interval=retry_interval,
            initial_deposit=initial_deposit,
            topup_deposit=topup_deposit
        )
    else:
        _http_client = http_client


def _check_init():
    assert _http_client, 'microraiden.requests has not been initialized. ' \
                         'Please call microraiden.requests.init() first.'


def request(method: str, url: str, **kwargs) -> Response:
    _check_init()
    return _http_client.request(method, url, **kwargs)


def head(url: str, **kwargs) -> Response:
    return request('HEAD', url, **kwargs)


def get(url: str, **kwargs) -> Response:
    return request('GET', url, **kwargs)


def post(url: str, **kwargs) -> Response:
    return request('POST', url, **kwargs)


def put(url: str, **kwargs) -> Response:
    return request('PUT', url, **kwargs)


def patch(url: str, **kwargs) -> Response:
    return request('PATCH', url, **kwargs)


def delete(url: str, **kwargs) -> Response:
    return request('DELETE', url, **kwargs)



class MicroRaidenException(Exception):
    pass


class InvalidBalanceAmount(MicroRaidenException):
    pass


class InvalidBalanceProof(MicroRaidenException):
    pass


class NoOpenChannel(MicroRaidenException):
    pass


class InsufficientConfirmations(MicroRaidenException):
    pass


class NoBalanceProofReceived(MicroRaidenException):
    pass


class StateFileException(MicroRaidenException):
    pass


class InvalidContractVersion(MicroRaidenException):
    pass


class StateContractAddrMismatch(StateFileException):
    pass


class StateReceiverAddrMismatch(StateFileException):
    pass


class StateFileLocked(StateFileException):
    pass


class InsecureStateFile(StateFileException):
    pass


class NetworkIdMismatch(StateFileException):
    pass



# import json
# import shutil
import sqlite3
import os
import logging
from eth_utils import is_address

from microraiden.utils import check_permission_safety

from microraiden.exceptions import (
    InsecureStateFile
)
from .channel import Channel, ChannelState

log = logging.getLogger(__name__)


def dict_factory(cursor, row):
    d = {}
    for idx, col in enumerate(cursor.description):
        d[col[0]] = row[idx]
    return d


DB_CREATION_SQL = """
CREATE TABLE `metadata` (
    `network_id`       INTEGER,
    `contract_address` CHAR(42),
    `receiver`         CHAR(42)
);
CREATE TABLE `syncstate` (
    `confirmed_head_number`   INTEGER,
    `confirmed_head_hash`     CHAR(66),
    `unconfirmed_head_number` INTEGER,
    `unconfirmed_head_hash`   CHAR(66)
);
-- deposit and balance have length of 78 to fit uint256
CREATE TABLE `channels` (
    `sender`            CHAR(42)        NOT NULL,
    `open_block_number` INTEGER         NOT NULL,
    `deposit`           DECIMAL(78,0)   NOT NULL,
    `balance`           DECIMAL(78,0)   NOT NULL,
    `last_signature`    CHAR(132),
    `settle_timeout`    INTEGER         NOT NULL,
    `mtime`             INTEGER         NOT NULL,
    `ctime`             INTEGER         NOT NULL,
    `state`             INTEGER         NOT NULL,
    `confirmed`         BOOL            NOT NULL,
    PRIMARY KEY (`sender`, `open_block_number`)
);
CREATE TABLE `topups` (
    `channel_rowid`     INTEGER,
    `txhash`            CHAR(66)        NOT NULL,
    `deposit`           DECIMAL(78,0)   NOT NULL,
    PRIMARY KEY (`channel_rowid`, `txhash`),
    FOREIGN KEY (`channel_rowid`) REFERENCES channels (rowid)
        ON DELETE CASCADE
);
INSERT INTO `metadata` VALUES (
    NULL,
    NULL,
    NULL
);
INSERT INTO `syncstate` VALUES (
    NULL,
    NULL,
    NULL,
    NULL
);
"""

UPDATE_METADATA_SQL = """
UPDATE `metadata` SET
    `network_id` = ?,
    `contract_address` = ?,
    `receiver` = ?;
"""

UPDATE_SYNCSTATE_SQL = {
    'confirmed_head_number': 'UPDATE `syncstate` SET `confirmed_head_number` = ?;',
    'confirmed_head_hash': 'UPDATE `syncstate` SET `confirmed_head_hash` = ?;',
    'unconfirmed_head_number': 'UPDATE `syncstate` SET `unconfirmed_head_number` = ?;',
    'unconfirmed_head_hash': 'UPDATE `syncstate` SET `unconfirmed_head_hash` = ?;'
}

ADD_CHANNEL_SQL = """
INSERT OR REPLACE INTO `channels` VALUES (
    ?,
    ?,
    ?,
    ?,
    ?,
    ?,
    ?,
    ?,
    ?,
    ?
)
"""

UPDATE_CHANNEL_SQL = """
UPDATE `channels` SET
    `deposit` = ?,
    `balance` = ?,
    `last_signature` = ?,
    `settle_timeout` = ?,
    `mtime` = ?,
    `state` = ?
WHERE `sender` = ? AND `open_block_number` = ?;
"""

DEL_CHANNEL_SQL = """
DELETE FROM `channels` WHERE `sender` = ? AND `open_block_number` = ?"""


class ChannelManagerState(object):
    """The part of the channel manager state that needs to persist."""

    def __init__(self, filename):
        self.filename = filename
        self.conn = sqlite3.connect(self.filename, isolation_level="EXCLUSIVE")
        self.conn.row_factory = dict_factory
        if filename not in (None, ':memory:'):
            os.chmod(filename, 0o600)

    def setup_db(self, network_id, contract_address, receiver):
        self.conn.executescript(DB_CREATION_SQL)
        self.conn.execute(UPDATE_METADATA_SQL, [network_id, contract_address, receiver])
        self.conn.commit()

    @property
    def contract_address(self):
        """The address of the channel manager contract."""
        c = self.conn.cursor()
        c.execute('SELECT `contract_address` FROM `metadata`;')
        contract_address = c.fetchone()['contract_address']
        assert c.fetchone() is None
        return contract_address

    @property
    def receiver(self):
        """The receiver address."""
        c = self.conn.cursor()
        c.execute('SELECT `receiver` FROM `metadata`;')
        receiver = c.fetchone()['receiver']
        assert c.fetchone() is None
        return receiver

    @property
    def network_id(self):
        """The receiver address."""
        c = self.conn.cursor()
        c.execute('SELECT `network_id` FROM `metadata`;')
        network_id = c.fetchone()['network_id']
        assert c.fetchone() is None
        return network_id

    @property
    def _sync_state(self):
        c = self.conn.cursor()
        c.execute('SELECT * FROM `syncstate`;')
        state = c.fetchone()
        assert c.fetchone() is None
        assert len(state) == 4
        return state

    @property
    def confirmed_head_number(self):
        """The number of the highest processed block considered to be final."""
        return self._sync_state['confirmed_head_number']

    @confirmed_head_number.setter
    def confirmed_head_number(self, value):
        self.update_sync_state(confirmed_head_number=value)

    @property
    def confirmed_head_hash(self):
        """The hash of the highest processed block considered to be final."""
        return self._sync_state['confirmed_head_hash']

    @confirmed_head_hash.setter
    def confirmed_head_hash(self, value):
        self.update_sync_state(confirmed_head_hash=value)

    @property
    def unconfirmed_head_number(self):
        """The number of the highest processed block considered to be not yet final."""
        return self._sync_state['unconfirmed_head_number']

    @unconfirmed_head_number.setter
    def unconfirmed_head_number(self, value):
        self.update_sync_state(unconfirmed_head_number=value)

    @property
    def unconfirmed_head_hash(self):
        """The hash of the highest processed block considered to be not yet final."""
        return self._sync_state['unconfirmed_head_hash']

    @unconfirmed_head_hash.setter
    def unconfirmed_head_hash(self, value):
        self.update_sync_state(unconfirmed_head_hash=value)

    def update_sync_state(
        self,
        confirmed_head_number=None,
        confirmed_head_hash=None,
        unconfirmed_head_number=None,
        unconfirmed_head_hash=None
    ):
        """Update block numbers and hashes of confirmed and unconfirmed head."""
        if confirmed_head_number is not None:
            sql = UPDATE_SYNCSTATE_SQL['confirmed_head_number']
            self.conn.execute(sql, [confirmed_head_number])
        if confirmed_head_hash is not None:
            sql = UPDATE_SYNCSTATE_SQL['confirmed_head_hash']
            self.conn.execute(sql, [confirmed_head_hash])
        if unconfirmed_head_number is not None:
            sql = UPDATE_SYNCSTATE_SQL['unconfirmed_head_number']
            self.conn.execute(sql, [unconfirmed_head_number])
        if unconfirmed_head_hash is not None:
            sql = UPDATE_SYNCSTATE_SQL['unconfirmed_head_hash']
            self.conn.execute(sql, [unconfirmed_head_hash])
        self.conn.commit()

    @property
    def n_channels(self):
        c = self.conn.cursor()
        c.execute('SELECT COUNT(*) as count FROM `channels`')
        return c.fetchone()['count']

    @property
    def n_open_channels(self):
        c = self.conn.cursor()
        c.execute('SELECT COUNT(*) as count FROM `channels` WHERE `state` = ?',
                  [ChannelState.OPEN.value])
        return c.fetchone()['count']

    def get_channels(self, confirmed=True):
        ret = dict()
        c = self.conn.cursor()
        c.execute('SELECT rowid, * FROM `channels` WHERE `confirmed` = ?', [confirmed])
        for result in c.fetchall():
            channel = self.result_to_channel(result)
            ret[result['sender'], result['open_block_number']] = channel
        return ret

    @property
    def channels(self):
        return self.get_channels(confirmed=True)

    @property
    def unconfirmed_channels(self):
        return self.get_channels(confirmed=False)

    @property
    def pending_channels(self):
        ret = dict()
        c = self.conn.cursor()
        c.execute('SELECT rowid, * FROM `channels` WHERE `state` = ?',
                  [ChannelState.CLOSE_PENDING.value])
        for result in c.fetchall():
            channel = self.result_to_channel(result)
            ret[result['sender'], result['open_block_number']] = channel
        return ret

    def result_to_channel(self, result):
        """Helper function to serialize one row of `channels` table into a channel object
        """
        channel = Channel(self.receiver, result['sender'],
                          result['deposit'],
                          result['open_block_number'])
        channel.balance = result['balance']
        channel.state = ChannelState(result['state'])
        channel.last_signature = result['last_signature']
        channel.settle_timeout = result['settle_timeout']
        channel.mtime = result['mtime']
        channel.ctime = result['ctime']
        channel.unconfirmed_topups = self.get_unconfirmed_topups(result['rowid'])
        channel.confirmed = result['confirmed']
        return channel

    def get_channel_rowid(self, sender: str, open_block_number: int):
        c = self.conn.cursor()
        result = c.execute(
            'SELECT rowid from `channels` WHERE sender = ? AND open_block_number = ?',
            [sender, open_block_number]
        )
        return result.fetchone()['rowid']

    def get_unconfirmed_topups(self, channel_rowid):
        c = self.conn.cursor()
        c.execute('SELECT * FROM topups WHERE channel_rowid = ?', [channel_rowid])
        return {result['txhash']: result['deposit'] for result in c.fetchall()}

    def set_channel(self, channel):
        self.add_channel(channel)

    def channel_exists(self, sender, open_block_number):
        c = self.conn.cursor()
        sql = 'SELECT 1 FROM `channels` WHERE `sender` = ? AND `open_block_number` == ?'
        c.execute(sql, [sender.lower(), open_block_number])
        result = c.fetchone()
        if result is None:
            return False
        elif result['1']:
            return True
        assert False

    def set_unconfirmed_topups(self, channel_rowid: int, topups: dict):
        assert channel_rowid is not None and isinstance(channel_rowid, int)
        self.conn.execute('DELETE FROM topups WHERE channel_rowid = ?', [channel_rowid])
        for txhash, deposit in topups.items():
            self.conn.execute('INSERT OR REPLACE INTO topups VALUES (?, ?, ?)',
                              [channel_rowid, txhash, str(deposit)])

    def add_channel(self, channel):
        assert channel.open_block_number > 0
        assert channel.state is not ChannelState.UNDEFINED
        assert is_address(channel.sender.lower())
        params = [
            channel.sender.lower(),
            channel.open_block_number,
            str(channel.deposit),
            str(channel.balance),
            channel.last_signature,
            channel.settle_timeout,
            channel.mtime,
            channel.ctime,
            channel.state.value,
            channel.confirmed
        ]
        self.conn.execute(ADD_CHANNEL_SQL, params)
        rowid = self.get_channel_rowid(channel.sender, channel.open_block_number)
        self.set_unconfirmed_topups(rowid, channel.unconfirmed_topups)
        self.conn.commit()

    def get_channel(self, sender, open_block_number):
        assert is_address(sender)
        assert open_block_number > 0
        # TODO unconfirmed topups
        c = self.conn.cursor()
        sql = 'SELECT rowid,* FROM `channels` WHERE `sender` = ? AND `open_block_number` = ?'
        c.execute(sql, [sender.lower(), open_block_number])
        result = c.fetchone()
        assert c.fetchone() is None
        return self.result_to_channel(result)

    def del_channel(self, sender: str, open_block_number: int):
        assert is_address(sender)
        assert open_block_number > 0
        assert self.channel_exists(sender, open_block_number)
        self.conn.execute(DEL_CHANNEL_SQL, [sender.lower(), open_block_number])
        self.conn.commit()

    @classmethod
    def load(cls, filename: str, check_permissions=True):
        """Load a previously stored state."""
        assert filename and isinstance(filename, str)
        if filename != ':memory:':
            if os.path.isfile(filename) is False:
                log.error("State file  %s doesn't exist" % filename)
                return None
            if check_permissions and not check_permission_safety(filename):
                raise InsecureStateFile(filename)
        ret = cls(filename)
        log.debug("loaded saved state. head_number=%s receiver=%s" %
                  (ret.confirmed_head_number, ret.receiver))
        # for sender, block in ret.channels.keys():
        #     log.debug("loaded channel info from the saved state sender=%s open_block=%s" %
        #               (sender, block))
        return ret

    def del_unconfirmed_channels(self):
        self.conn.execute('DELETE FROM `channels` WHERE `confirmed` = 0')
        self.conn.commit()

    def set_channel_state(self, sender: str, open_block_number: int, state: ChannelState):
        self.conn.execute('UPDATE `channels` SET `state` = ?'
                          'WHERE `sender` = ? AND `open_block_number` = ?',
                          [state, sender, open_block_number])



from .channel import Channel
from .client import Client
from .http_client import HTTPClient
from .default_http_client import DefaultHTTPClient

__all__ = [
    Channel,
    Client,
    HTTPClient,
    DefaultHTTPClient
]



import pytest

from eth_utils import (
    encode_hex,
)
from ethereum.tester import TransactionFailed
from web3 import Web3
from web3.exceptions import BadFunctionCallOutput

from microraiden import Client
from microraiden.channel_manager import ChannelManager
from microraiden.close_all_channels import close_open_channels


def test_close_simple(
        client: Client,
        channel_manager: ChannelManager,
        web3: Web3,
        wait_for_blocks
):
    sender = client.context.address
    receiver = channel_manager.receiver

    channel = client.open_channel(receiver, 10)
    wait_for_blocks(channel_manager.n_confirmations + 1)
    channel_manager.register_payment(sender, channel.block, 2,
                                     encode_hex(channel.create_transfer(2)))

    channel_manager.stop()  # don't update state from this point on
    channel_manager.join()
    state = channel_manager.state
    tx_count_before = web3.eth.getTransactionCount(receiver)
    close_open_channels(
        channel_manager.private_key,
        state,
        channel_manager.channel_manager_contract,
        10,
        wait=lambda: wait_for_blocks(1)
    )
    tx_count_after = web3.eth.getTransactionCount(receiver)
    assert tx_count_after == tx_count_before + 1

    with pytest.raises((BadFunctionCallOutput, TransactionFailed)):
        channel_id = (channel.sender, channel.receiver, channel.block)
        channel_manager.channel_manager_contract.call().getChannelInfo(*channel_id)

    wait_for_blocks(1)


def test_close_topup(
        client: Client,
        channel_manager: ChannelManager,
        web3: Web3,
        wait_for_blocks
):
    sender = client.context.address
    receiver = channel_manager.receiver

    channel = client.open_channel(receiver, 10)
    wait_for_blocks(channel_manager.n_confirmations + 1)
    channel.topup(5)
    wait_for_blocks(channel_manager.n_confirmations + 1)
    channel_manager.register_payment(sender, channel.block, 12,
                                     encode_hex(channel.create_transfer(12)))

    channel_manager.stop()  # don't update state from this point on
    channel_manager.join()
    state = channel_manager.state
    tx_count_before = web3.eth.getTransactionCount(receiver)
    close_open_channels(
        channel_manager.private_key,
        state,
        channel_manager.channel_manager_contract,
        10,
        wait=lambda: wait_for_blocks(1)
    )
    tx_count_after = web3.eth.getTransactionCount(receiver)
    assert tx_count_after == tx_count_before + 1

    with pytest.raises((BadFunctionCallOutput, TransactionFailed)):
        channel_id = (channel.sender, channel.receiver, channel.block)
        channel_manager.channel_manager_contract.call().getChannelInfo(*channel_id)
    wait_for_blocks(1)


def test_close_valid_close(
        client: Client,
        channel_manager: ChannelManager,
        web3: Web3,
        wait_for_blocks
):
    sender = client.context.address
    receiver = channel_manager.receiver

    channel = client.open_channel(receiver, 10)
    wait_for_blocks(channel_manager.n_confirmations + 1)
    channel_manager.register_payment(sender, channel.block, 2,
                                     encode_hex(channel.create_transfer(2)))
    channel.close()

    channel_manager.stop()  # don't update state from this point on
    channel_manager.join()
    state = channel_manager.state

    tx_count_before = web3.eth.getTransactionCount(receiver)
    close_open_channels(
        channel_manager.private_key,
        state,
        channel_manager.channel_manager_contract,
        10,
        wait=lambda: wait_for_blocks(1)
    )
    tx_count_after = web3.eth.getTransactionCount(receiver)
    assert tx_count_after == tx_count_before + 1

    with pytest.raises((BadFunctionCallOutput, TransactionFailed)):
        channel_id = (channel.sender, channel.receiver, channel.block)
        channel_manager.channel_manager_contract.call().getChannelInfo(*channel_id)
    wait_for_blocks(1)


def test_close_invalid_close(
        client: Client,
        channel_manager: ChannelManager,
        web3: Web3,
        wait_for_blocks
):
    sender = client.context.address
    receiver = channel_manager.receiver

    channel = client.open_channel(receiver, 10)
    wait_for_blocks(channel_manager.n_confirmations + 1)
    channel_manager.register_payment(sender, channel.block, 2,
                                     encode_hex(channel.create_transfer(2)))
    # cheat
    channel.update_balance(0)
    channel.create_transfer(1)
    channel.close()

    channel_manager.stop()  # don't update state from this point on
    channel_manager.join()
    state = channel_manager.state

    tx_count_before = web3.eth.getTransactionCount(receiver)
    close_open_channels(
        channel_manager.private_key,
        state,
        channel_manager.channel_manager_contract,
        10,
        wait=lambda: wait_for_blocks(1)
    )
    tx_count_after = web3.eth.getTransactionCount(receiver)
    assert tx_count_after == tx_count_before + 1

    with pytest.raises((BadFunctionCallOutput, TransactionFailed)):
        channel_id = (channel.sender, channel.receiver, channel.block)
        channel_manager.channel_manager_contract.call().getChannelInfo(*channel_id)
    wait_for_blocks(1)


def test_close_settled(
        client: Client,
        channel_manager: ChannelManager,
        web3: Web3,
        wait_for_blocks
):
    sender = client.context.address
    receiver = channel_manager.receiver

    channel = client.open_channel(receiver, 10)
    wait_for_blocks(channel_manager.n_confirmations + 1)
    channel_manager.register_payment(sender, channel.block, 2,
                                     encode_hex(channel.create_transfer(2)))
    receiver_sig = channel_manager.sign_close(sender, channel.block, 2)
    channel.close_cooperatively(receiver_sig)
    wait_for_blocks(channel_manager.n_confirmations + 1)

    channel_manager.stop()  # don't update state from this point on
    channel_manager.join()
    state = channel_manager.state

    tx_count_before = web3.eth.getTransactionCount(receiver)
    close_open_channels(
        channel_manager.private_key,
        state,
        channel_manager.channel_manager_contract,
        10,
        wait=lambda: wait_for_blocks(1)
    )
    tx_count_after = web3.eth.getTransactionCount(receiver)
    assert tx_count_after == tx_count_before
    wait_for_blocks(1)



import click

from microraiden.proxy.content import (
    PaywalledProxyUrl
)
from microraiden.click_helpers import main, pass_app
from microraiden.config import TKN_DECIMALS


@main.command()
@click.option(
    '--host',
    default='localhost',
    help='Address of the proxy'
)
@click.option(
    '--port',
    default=5000,
    help='Port of the proxy'
)
@pass_app
def start(app, host, port):
    app.add_content(PaywalledProxyUrl(".*",
                                      1 * TKN_DECIMALS,
                                      "http://en.wikipedia.org/",
                                      [r"wiki/.*"]))
    app.run(host=host, port=port, debug=True)
    app.join()


if __name__ == '__main__':
    import logging
    from gevent import monkey
    monkey.patch_all()
    logging.basicConfig(level=logging.DEBUG)
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("blockchain").setLevel(logging.DEBUG)
    logging.getLogger("channel_manager").setLevel(logging.DEBUG)
    main()



import time
import uuid
from flask_httpauth import HTTPBasicAuth
from flask import g
from flask_restful import Resource

auth = HTTPBasicAuth()


class TokenAccess:
    def __init__(self, user: str):
        self.token = str(uuid.uuid1())
        self.time_created = time.time()
        self.time_accessed = time.time()
        self.user = user


class UsersDB:
    def __init__(self):
        self.users = {}
        self.tokens = {}
        self.token_expiry_seconds = 600  # token is valid for 10 minutes

    def add_user(self, user: str, password: str):
        self.users[user] = password

    def del_user(self, user: str):
        self.users.pop(user)

    def authorize(self, user_or_token: str, password: str):
        """Authorize user using token or username/password combination"""
        g.user = user_or_token
        token_record = self.verify_token(user_or_token)
        if token_record is not None:
            token_record.time_accessed = time.time()
            return True
        else:
            self.tokens.pop(user_or_token, '')
        if user_or_token not in self.users:
            return False
        return self.users[user_or_token] == password

    def verify_token(self, token: str):
        """Verify if the token is valid and not expired"""
        token_record = self.tokens.get(token, None)
        if token_record is None:
            return None
        t_diff = time.time() - token_record.time_accessed
        assert t_diff >= 0
        if t_diff > self.token_expiry_seconds:
            return None
        return token_record

    def remove_token(self, token: str):
        del self.tokens[token]

    def get_token(self, user: str):
        token_record = TokenAccess(user)
        self.tokens[token_record.token] = token_record
        return token_record.token


userDB = UsersDB()


# used by flask to process http auth requests
@auth.verify_password
def verify_password(username, password):
    return userDB.authorize(username, password)


# two resources for managing the login
# exported as /login and /logout
class ChannelManagementLogin(Resource):
    @auth.login_required
    def get(self):
        token = userDB.get_token(g.user)
        return {'token': token}, 200


class ChannelManagementLogout(Resource):
    @auth.login_required
    def get(self):
        userDB.remove_token(g.user)
        return "OK", 200



from .variables import *    # noqa
from .client import *       # noqa
from .http_client import *  # noqa
from .web3 import *         # noqa
from .channel_manager import *  # noqa
from .proxy import *        # noqa
from .accounts import *     # noqa



import logging
import gevent

log = logging.getLogger(__name__)


def test_big_deposit(channel_manager, client, receiver_address, wait_for_blocks):
    """Test if deposit of size bigger than int64 causes havoc when storing the state."""
    BIG_DEPOSIT = 10000000000000000000
    blockchain = channel_manager.blockchain

    channel_manager.wait_sync()
    channel = client.open_channel(receiver_address, BIG_DEPOSIT)
    wait_for_blocks(channel_manager.blockchain.n_confirmations)
    gevent.sleep(blockchain.poll_interval)

    channel_rec = channel_manager.channels[channel.sender, channel.block]
    assert channel_rec.deposit == BIG_DEPOSIT



import pytest
from microraiden.channel_manager import (
    Channel,
    ChannelState,
    ChannelManagerState
)


CONTRACT_ADDRESS = '0xaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'
RECEIVER_ADDRESS = '0xbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb'
SENDER_ADDRESS = '0xcccccccccccccccccccccccccccccccccccccccc'
NETWORK_ID = 123
BLOCK_HASH = '0xaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'
SIG = '0x' + 'bb' * 130


@pytest.fixture()
def state(tmpdir):
    db = tmpdir.join("state.db")
    state = ChannelManagerState(db.strpath)
    state.setup_db(
        NETWORK_ID,
        CONTRACT_ADDRESS,
        RECEIVER_ADDRESS
    )
    return state


def test_creation(state):
    assert state.receiver == RECEIVER_ADDRESS
    assert state.contract_address == CONTRACT_ADDRESS
    assert state.network_id == NETWORK_ID

    assert state.confirmed_head_number is None
    assert state.confirmed_head_hash is None
    assert state.unconfirmed_head_number is None
    assert state.unconfirmed_head_hash is None
    assert state.n_channels == 0
    assert state.n_open_channels == 0

    state_loaded = ChannelManagerState.load(state.filename, check_permissions=False)
    assert state_loaded.confirmed_head_number is None
    assert state_loaded.confirmed_head_hash is None
    assert state_loaded.unconfirmed_head_number is None
    assert state_loaded.unconfirmed_head_hash is None
    assert state_loaded.n_channels == 0
    assert state_loaded.n_open_channels == 0


def test_sync_state(state):
    state.update_sync_state(confirmed_head_hash=BLOCK_HASH)
    assert state.confirmed_head_hash == BLOCK_HASH
    state.update_sync_state(confirmed_head_number=123)
    assert state.confirmed_head_number == 123
    state.update_sync_state(unconfirmed_head_hash=BLOCK_HASH)
    assert state.unconfirmed_head_hash == BLOCK_HASH
    state.update_sync_state(unconfirmed_head_number=321)
    assert state.unconfirmed_head_number == 321

    state_loaded = ChannelManagerState.load(state.filename, check_permissions=False)
    assert state_loaded.confirmed_head_hash == BLOCK_HASH
    assert state_loaded.confirmed_head_number == 123
    assert state_loaded.unconfirmed_head_hash == BLOCK_HASH
    assert state_loaded.unconfirmed_head_number == 321


def test_adding_channel(state):
    channel = Channel(RECEIVER_ADDRESS, SENDER_ADDRESS, 100, 123)
    channel.balance = 50
    channel.state = ChannelState.OPEN
    assert not state.channel_exists(channel.sender, channel.open_block_number)
    state.add_channel(channel)
    assert state.channel_exists(channel.sender, channel.open_block_number)
    channel_retrieved = state.get_channel(channel.sender, channel.open_block_number)
    assert channel_retrieved.sender == SENDER_ADDRESS
    assert channel_retrieved.receiver == RECEIVER_ADDRESS
    assert channel_retrieved.open_block_number == 123
    assert channel_retrieved.deposit == 100
    assert channel_retrieved.balance == 50
    assert channel_retrieved.last_signature is None
    assert channel_retrieved.is_closed is False
    assert channel_retrieved.ctime == channel.ctime
    assert channel_retrieved.mtime == channel.mtime


def test_updating_channel(state):
    channel = Channel(RECEIVER_ADDRESS, SENDER_ADDRESS, 100, 123)
    channel.balance = 50
    channel.state = ChannelState.OPEN
    state.add_channel(channel)
    channel.deposit = 200
    channel.balance = 100
    channel.last_signature = SIG
    channel.is_closed = True
    channel.mtime = channel.mtime + 1
    state.set_channel(channel)
    channel_retrieved = state.get_channel(channel.sender, channel.open_block_number)
    assert channel_retrieved.sender == SENDER_ADDRESS
    assert channel_retrieved.receiver == RECEIVER_ADDRESS
    assert channel_retrieved.open_block_number == 123
    assert channel_retrieved.deposit == 200
    assert channel_retrieved.balance == 100
    assert channel_retrieved.last_signature == SIG
    assert channel_retrieved.is_closed is True
    assert channel_retrieved.ctime == channel.ctime
    assert channel_retrieved.mtime == channel.mtime
